import networkx as nx
import pytest

from region.p_regions.exact import PRegionsExact
from region.tests.util import region_list_from_array, compare_region_lists
from region.util import dataframe_to_dict

from region.p_regions.tests.data import (
    adj,
    neighbors_dict,
    gdf,
    graph,
    w,
    attr,
    attr_dict,
    attr_str,
    double_attr_str,
    double_attr,
    double_attr_dict,
    optimal_clustering,
)


@pytest.fixture(params=["flow", "order", "tree"])
def method(request):
    return request.param


# ### TESTS WITH SCALAR attr ##################################################
# test with csr_matrix
def test_scipy_sparse_matrix(method):
    cluster_object = PRegionsExact()
    cluster_object.fit_from_scipy_sparse_matrix(adj, attr, n_regions=2, method=method)
    obtained = region_list_from_array(cluster_object.labels_)
    compare_region_lists(obtained, optimal_clustering)


# tests with a GeoDataFrame as areas argument
def test_geodataframe(method):
    cluster_object = PRegionsExact()
    cluster_object.fit_from_geodataframe(gdf, attr_str, n_regions=2, method=method)
    result = region_list_from_array(cluster_object.labels_)
    compare_region_lists(result, optimal_clustering)


# tests with a dict as areas argument
def test_dict(method):
    value_dict = dataframe_to_dict(gdf, attr_str)
    cluster_object = PRegionsExact()
    cluster_object.fit_from_dict(neighbors_dict, value_dict, n_regions=2, method=method)
    result = region_list_from_array(cluster_object.labels_)
    compare_region_lists(result, optimal_clustering)


# tests with Graph
# ... with dicts as attr and spatially_extensive_attr
def test_graph_dict_basic(method):
    cluster_object = PRegionsExact()
    cluster_object.fit_from_networkx(graph, attr_dict, n_regions=2, method=method)
    result = region_list_from_array(cluster_object.labels_)
    compare_region_lists(result, optimal_clustering)


# ... with strings as attr and spatially_extensive_attr
def test_graph_str_basic(method):
    nx.set_node_attributes(graph, attr_dict, attr_str)
    cluster_object = PRegionsExact()
    cluster_object.fit_from_networkx(graph, attr_str, n_regions=2, method=method)
    result = region_list_from_array(cluster_object.labels_)
    compare_region_lists(result, optimal_clustering)


# test with W
def test_w_basic(method):
    cluster_object = PRegionsExact()
    cluster_object.fit_from_w(w, attr, n_regions=2, method=method)
    result = region_list_from_array(cluster_object.labels_)
    compare_region_lists(result, optimal_clustering)


# ### TESTS WITH NON-SCALAR attr AND spatially_extensive_attr #################
# test with csr_matrix
def test_scipy_sparse_matrix_multi_attr(method):
    cluster_object = PRegionsExact()
    cluster_object.fit_from_scipy_sparse_matrix(
        adj, double_attr, n_regions=2, method=method
    )
    obtained = region_list_from_array(cluster_object.labels_)
    compare_region_lists(obtained, optimal_clustering)


# tests with a GeoDataFrame
def test_geodataframe_multi_attr(method):
    cluster_object = PRegionsExact()
    cluster_object.fit_from_geodataframe(
        gdf, double_attr_str, n_regions=2, method=method
    )
    obtained = region_list_from_array(cluster_object.labels_)
    compare_region_lists(obtained, optimal_clustering)


# tests with a dict as areas argument
def test_dict_multi_attr(method):
    cluster_object = PRegionsExact()
    cluster_object.fit_from_dict(
        neighbors_dict, double_attr_dict, n_regions=2, method=method
    )
    obtained = region_list_from_array(cluster_object.labels_)
    compare_region_lists(obtained, optimal_clustering)


# tests with Graph
# ... with dicts as attr and spatially_extensive_attr
def test_graph_dict_multi_attr(method):
    cluster_object = PRegionsExact()
    cluster_object.fit_from_networkx(
        graph, double_attr_dict, n_regions=2, method=method
    )
    result = region_list_from_array(cluster_object.labels_)
    compare_region_lists(result, optimal_clustering)


# ... with strings as attr and spatially_extensive_attr
def test_graph_str_multi_attr(method):
    nx.set_node_attributes(graph, attr_dict, attr_str)
    cluster_object = PRegionsExact()
    cluster_object.fit_from_networkx(graph, double_attr_str, n_regions=2, method=method)
    result = region_list_from_array(cluster_object.labels_)
    compare_region_lists(result, optimal_clustering)


# test with W
def test_w_multi_attr(method):
    cluster_object = PRegionsExact()
    cluster_object.fit_from_w(w, double_attr, n_regions=2, method=method)
    result = region_list_from_array(cluster_object.labels_)
    compare_region_lists(result, optimal_clustering)



from libpysal.io import geotable
from libpysal import examples, weights
import numpy as np
from sklearn.metrics import pairwise as skm
from region.skater.skater import Spanning_Forest

import types
import os

TESTDIR = os.path.dirname(os.path.abspath(__file__))


df = geotable.read_files(examples.get_path('south.shp'))
data = df[df.filter(like='90').columns.tolist()
               + df.filter(like='89').columns.tolist()].values
data_c = (data - data.mean(axis=0)) / data.std(axis=0)
W = weights.Queen.from_dataframe(df)


def test_init():
    default = Spanning_Forest()
    assert default.metric == skm.manhattan_distances
    assert default.center == np.mean
    assert default.reduction == np.sum
    change = Spanning_Forest(dissimilarity=skm.euclidean_distances,
                             center=np.median, reduction=np.max)
    assert change.metric == skm.euclidean_distances
    assert change.center == np.median
    assert change.reduction == np.max
    
    sym = Spanning_Forest(affinity=skm.cosine_similarity)
    assert isinstance(sym.metric, types.LambdaType)
    test_distance = -np.log(skm.cosine_similarity(data[:2,]))
    comparator = sym.metric(data[:2,])
    np.testing.assert_allclose(test_distance, comparator)


def test_run():
    result = Spanning_Forest().fit(5, W, data_c, quorum=50)
    
    #hmm... numbering is arbitrary... just check that they run for now
    #np.testing.assert_array_equal(self.south_5_q100, result.current_labels_)
    result2 = Spanning_Forest().fit(np.inf, W, data_c, quorum=20)
    
    #np.testing.assert_array_equal(self.south_inf_q20, result2.current_labels_)




import collections
import numbers
from math import floor, log10

from geopandas import GeoDataFrame
import numpy as np
import networkx as nx
from pulp import LpProblem, LpMinimize, LpVariable, LpInteger, lpSum

from region.csgraph_utils import neighbors
from region.util import array_from_df_col, array_from_dict_values,\
    array_from_graph_or_dict, check_solver, copy_func, get_metric_function,\
    get_solver_instance, raise_distance_metric_not_set,\
    scipy_sparse_matrix_from_dict, scipy_sparse_matrix_from_w, w_from_gdf


class MaxPRegionsExact:
    """
    A class for solving the max-p-regions problem by transforming it into a
    mixed-integer-programming problem (MIP) as described in [DAR2012]_.

    Attributes
    ----------
    labels_ : dict
        Each key is an area and each value the region it has been assigned to.
    """

    def __init__(self):
        self.labels_ = None
        self.solver = None
        self.metric = raise_distance_metric_not_set

    def fit_from_scipy_sparse_matrix(self,
                                     adj,
                                     attr,
                                     spatially_extensive_attr,
                                     threshold,
                                     solver="cbc",
                                     metric="euclidean"):
        """
        Solve the max-p-regions problem as MIP as described in [DAR2012]_.

        The resulting region labels are assigned to the instance's
        :attr:`labels_` attribute.

        Parameters
        ----------
        adj : class:`scipy.sparse.csr_matrix`
            Adjacency matrix representing the areas' contiguity relation.
        attr : :class:`numpy.ndarray`
            Array (number of areas x number of attributes) of areas' attributes
            relevant to clustering.
        spatially_extensive_attr : :class:`numpy.ndarray`
            Array (number of areas x number of attributes) of areas' attributes
            relevant to ensuring the threshold condition.
        threshold : numbers.Real or :class:`numpy.ndarray`
            The lower bound for a region's sum of spatially extensive
            attributes. The argument's type is numbers.Real if there is only
            one spatially extensive attribute per area, otherwise it is a
            one-dimensional array with as many entries as there are spatially
            extensive attributes per area.
        solver : {"cbc", "cplex", "glpk", "gurobi"}, default: "cbc"
            The solver to use. Unless the default solver is used, the user has
            to make sure that the specified solver is installed.

            * "cbc" - the Cbc (Coin-or branch and cut) solver
            * "cplex" - the CPLEX solver
            * "glpk" - the GLPK (GNU Linear Programming Kit) solver
            * "gurobi" - the Gurobi Optimizer

        metric : str or function, default: "euclidean"
            See the `metric` argument in
            :func:`region.util.get_metric_function`.
        """
        self.metric = get_metric_function(metric)
        check_solver(solver)

        prob = LpProblem("Max-p-Regions", LpMinimize)

        # Parameters of the optimization problem
        n_areas = adj.shape[0]
        I = list(range(n_areas))  # index for areas
        II = [(i, j) for i in I for j in I]
        II_upper_triangle = [(i, j) for i, j in II if i < j]
        # index of potential regions, called k in [DAR2012]_:
        K = range(n_areas)
        # index of contiguity order, called c in [DAR2012]_:
        O = range(n_areas)
        d = {(i, j): self.metric(attr[i].reshape(1, -1), attr[j].reshape(
            1, -1))
             for i, j in II_upper_triangle}
        h = 1 + floor(log10(sum(d[(i, j)] for i, j in II_upper_triangle)))

        # Decision variables
        t = LpVariable.dicts(
            "t", ((i, j) for i, j in II_upper_triangle),
            lowBound=0,
            upBound=1,
            cat=LpInteger)
        x = LpVariable.dicts(
            "x", ((i, k, o) for i in I for k in K for o in O),
            lowBound=0,
            upBound=1,
            cat=LpInteger)

        # Objective function
        # (1) in Duque et al. (2012): "The Max-p-Regions Problem"
        prob += -10**h * lpSum(x[i, k, 0] for k in K for i in I) \
            + lpSum(d[i, j] * t[i, j] for i, j in II_upper_triangle)

        # Constraints
        # (2) in Duque et al. (2012): "The Max-p-Regions Problem"
        for k in K:
            prob += lpSum(x[i, k, 0] for i in I) <= 1
        # (3) in Duque et al. (2012): "The Max-p-Regions Problem"
        for i in I:
            prob += lpSum(x[i, k, o] for k in K for o in O) == 1
        # (4) in Duque et al. (2012): "The Max-p-Regions Problem"
        for i in I:
            for k in K:
                for o in range(1, len(O)):
                    prob += x[i, k, o] <= lpSum(x[j, k, o - 1]
                                                for j in neighbors(adj, i))
        # (5) in Duque et al. (2012): "The Max-p-Regions Problem"
        if isinstance(spatially_extensive_attr[I[0]], numbers.Real):
            for k in K:
                lhs = lpSum(x[i, k, o] * spatially_extensive_attr[i] for i in I
                            for o in O)
                prob += lhs >= threshold * lpSum(x[i, k, 0] for i in I)
        elif isinstance(spatially_extensive_attr[I[0]], collections.Iterable):
            for el in range(len(spatially_extensive_attr[I[0]])):
                for k in K:
                    lhs = lpSum(x[i, k, o] * spatially_extensive_attr[i][el]
                                for i in I for o in O)
                    if isinstance(threshold, numbers.Real):
                        rhs = threshold * lpSum(x[i, k, 0] for i in I)
                        prob += lhs >= rhs
                    elif isinstance(threshold, np.ndarray):
                        rhs = threshold[el] * lpSum(x[i, k, 0] for i in I)
                        prob += lhs >= rhs
        # (6) in Duque et al. (2012): "The Max-p-Regions Problem"
        for i, j in II_upper_triangle:
            for k in K:
                prob += t[i, j] >= \
                        lpSum(x[i, k, o] + x[j, k, o] for o in O) - 1
        # (7) in Duque et al. (2012): "The Max-p-Regions Problem"
        # already in LpVariable-definition
        # (8) in Duque et al. (2012): "The Max-p-Regions Problem"
        # already in LpVariable-definition

        # additional constraint for speedup (p. 405 in [DAR2012]_)
        for o in O:
            prob += x[I[0], K[0], o] == (1 if o == 0 else 0)

        # Solve the optimization problem
        solver = get_solver_instance(solver)
        prob.solve(solver=solver)
        result = np.zeros(n_areas)
        for i in I:
            for k in K:
                for o in O:
                    if x[i, k, o].varValue == 1:
                        result[i] = k
        self.labels_ = result
        self.solver = solver

    fit = copy_func(fit_from_scipy_sparse_matrix)
    fit.__doc__ = "Alias for :meth:`fit_from_scipy_sparse_matrix`.\n\n" \
                  + fit_from_scipy_sparse_matrix.__doc__

    def fit_from_dict(self,
                      neighbors_dict,
                      attr,
                      spatially_extensive_attr,
                      threshold,
                      solver="cbc",
                      metric="euclidean"):
        """
        Alternative API for :meth:`fit_from_scipy_sparse_matrix`.

        Parameters
        ----------
        neighbors_dict : dict
            Each key represents an area and each value is an iterable of
            neighbors of this area.
        attr : dict
            A dict with the same keys as `neighbors_dict` and values
            representing the attributes for calculating homo-/heterogeneity. A
            value can be scalar (e.g. `float` or `int`) or a
            :class:`numpy.ndarray`.
        spatially_extensive_attr : dict
            A dict with the same keys as `neighbors_dict` and values
            representing the spatially extensive attribute (scalar or iterable
            of scalars). In the max-p-regions problem each region's sum of
            spatially extensive attributes must be greater than a specified
            threshold. In case of iterables of scalars as dict-values all
            elements of the iterable have to fulfill the condition.
        threshold : numbers.Real or :class:`numpy.ndarray`
            Refer to the corresponding argument in
            :meth:`fit_from_scipy_sparse_matrix`.
        solver : {"cbc", "cplex", "glpk", "gurobi"}, default: "cbc"
            Refer to the corresponding argument in
            :meth:`fit_from_scipy_sparse_matrix`.
        metric : str or function, default: "euclidean"
            Refer to the corresponding argument in
            :meth:`fit_from_scipy_sparse_matrix`.
        """
        if not isinstance(neighbors_dict, dict):
            raise ValueError("The neighbors_dict argument must be dict.")

        not_same_dict_keys_msg = "The {} argument has to be of type dict " \
                                 "with the same keys as neighbors_dict."

        if not isinstance(attr, dict) or attr.keys() != neighbors_dict.keys():
            raise ValueError(not_same_dict_keys_msg.format("attr"))

        if not isinstance(spatially_extensive_attr, dict) or \
                spatially_extensive_attr.keys() != neighbors_dict.keys():
            raise ValueError(
                not_same_dict_keys_msg.format(spatially_extensive_attr))

        adj = scipy_sparse_matrix_from_dict(neighbors_dict)
        attr_arr = array_from_dict_values(attr)
        spat_ext_attr_arr = array_from_dict_values(spatially_extensive_attr)
        self.fit_from_scipy_sparse_matrix(
            adj,
            attr_arr,
            spat_ext_attr_arr,
            threshold=threshold,
            solver=solver,
            metric=metric)

    def fit_from_geodataframe(self,
                              gdf,
                              attr,
                              spatially_extensive_attr,
                              threshold,
                              solver="cbc",
                              metric="euclidean",
                              contiguity="rook"):
        """
        Alternative API for :meth:`fit_from_scipy_sparse_matrix`.

        Parameters
        ----------
        gdf : GeoDataFrame

        attr : str or list
            The clustering criteria (columns of the GeoDataFrame `gdf`) are
            specified as string (for one column) or list of strings (for
            multiple columns).
        spatially_extensive_attr : str or list
            The name (`str`) or names (`list` of strings) of column(s) in `gdf`
            containing the spatially extensive attributes.
        threshold : numbers.Real or :class:`numpy.ndarray`
            Refer to the corresponding argument in
            :meth:`fit_from_scipy_sparse_matrix`.
        solver : {"cbc", "cplex", "glpk", "gurobi"}, default: "cbc"
            Refer to the corresponding argument in
            :meth:`fit_from_scipy_sparse_matrix`.
        metric : str or function, default: "euclidean"
            Refer to the corresponding argument in
            :meth:`fit_from_scipy_sparse_matrix`.
        contiguity : {"rook", "queen"}, default: "rook"
            Defines the contiguity relationship between areas. Possible
            contiguity definitions are:

            * "rook" - Rook contiguity.
            * "queen" - Queen contiguity.
        """
        w = w_from_gdf(gdf, contiguity)
        attr = array_from_df_col(gdf, attr)
        spat_ext_attr = array_from_df_col(gdf, spatially_extensive_attr)

        self.fit_from_w(
            w,
            attr,
            spat_ext_attr,
            threshold=threshold,
            solver=solver,
            metric=metric)

    def fit_from_networkx(self,
                          graph,
                          attr,
                          spatially_extensive_attr,
                          threshold,
                          solver="cbc",
                          metric="euclidean"):
        """
        Alternative API for :meth:`fit_from_scipy_sparse_matrix`.

        Parameters
        ----------
        graph : `networkx.Graph`

        attr : str, list or dict
            If the clustering criteria are present in the networkx.Graph
            `graph` as node attributes, then they can be specified as a string
            (for one criterion) or as a list of strings (for multiple
            criteria).
            Alternatively, a dict can be used with each key being a node of the
            networkx.Graph `graph` and each value being the corresponding
            clustering criterion (a scalar (e.g. `float` or `int`) or a
            :class:`numpy.ndarray`).
            If there are no clustering criteria present in the networkx.Graph
            `graph` as node attributes, then a dictionary must be used for this
            argument. Refer to the corresponding argument in
            :meth:`fit_from_dict` for more details about the expected dict.
        spatially_extensive_attr : str, list or dict
            If the spatially extensive attribute is present in the
            networkx.Graph `graph` as node attributes, then they can be
            specified as a string (for one attribute) or as a list of
            strings (for multiple attributes).
            Alternatively, a dict can be used with each key being a node of the
            networkx.Graph `graph` and each value being the corresponding
            spatially extensive attribute (a scalar (e.g. `float` or `int`) or
            a :class:`numpy.ndarray`).
            If there are no spatially extensive attributes present in the
            networkx.Graph `graph` as node attributes, then a dictionary must
            be used for this argument. Refer to the corresponding argument in
            :meth:`fit_from_dict` for more details about the expected dict.
        threshold : numbers.Real or :class:`numpy.ndarray`
            Refer to the corresponding argument in
            :meth:`fit_from_scipy_sparse_matrix`.
        solver : {"cbc", "cplex", "glpk", "gurobi"}, default: "cbc"
            Refer to the corresponding argument in
            :meth:`fit_from_scipy_sparse_matrix`.
        metric : str or function, default: "euclidean"
            Refer to the corresponding argument in
            :meth:`fit_from_scipy_sparse_matrix`.
        """
        adj = nx.to_scipy_sparse_matrix(graph)
        attr = array_from_graph_or_dict(graph, attr)
        sp_ext_attr = array_from_graph_or_dict(graph, spatially_extensive_attr)
        self.fit_from_scipy_sparse_matrix(
            adj,
            attr,
            sp_ext_attr,
            threshold=threshold,
            solver=solver,
            metric=metric)

    def fit_from_w(self,
                   w,
                   attr,
                   spatially_extensive_attr,
                   threshold,
                   solver="cbc",
                   metric="euclidean"):
        """
        Alternative API for :meth:`fit_from_scipy_sparse_matrix`.

        Parameters
        ----------
        w : libpysal.weights.W
            W object representing the areas' contiguity relation.
        attr : :class:`numpy.ndarray`
            See the corresponding argument in
            :meth:`fit_from_scipy_sparse_matrix`.
        spatially_extensive_attr : :class:`numpy.ndarray`
            See the corresponding argument in
            :meth:`fit_from_scipy_sparse_matrix`.
        threshold : numbers.Real or :class:`numpy.ndarray`
            Refer to the corresponding argument in
            :meth:`fit_from_scipy_sparse_matrix`.
        solver : {"cbc", "cplex", "glpk", "gurobi"}, default: "cbc"
            Refer to the corresponding argument in
            :meth:`fit_from_scipy_sparse_matrix`.
        metric : str or function, default: "euclidean"
            Refer to the corresponding argument in
            :meth:`fit_from_scipy_sparse_matrix`.
        """
        adj = scipy_sparse_matrix_from_w(w)
        self.fit_from_scipy_sparse_matrix(
            adj,
            attr,
            spatially_extensive_attr,
            threshold=threshold,
            solver=solver,
            metric=metric)



"""
Utility functions for graph-related operations with sparse adjacency matrices
(scipy.sparse.csr_matrix) using and supplementing scipy's
[compressed sparse graph routines](
https://docs.scipy.org/doc/scipy/reference/sparse.csgraph.html).
"""
from scipy.sparse import csgraph as csg
from scipy.sparse import csr_matrix
import numpy as np


def is_connected(adj):
    """
    Parameters
    ----------
    adj : :class:`scipy.sparse.csr_matrix`
        Adjacency matrix.

    Returns
    -------
    connected : `bool`
        `True` if graph defined by adjecency matrix `adj` is connected.
        `False` otherwise.

    Examples
    --------
    >>> import numpy as np
    >>> from scipy.sparse import csr_matrix
    >>> connected = csr_matrix(np.array([[0, 1],
    ...                                  [1, 0]]))
    >>> is_connected(connected)
    True
    >>> disconnected = csr_matrix(np.array([[0, 0],
    ...                                     [0, 0]]))
    >>> is_connected(disconnected)
    False
    """
    n_connected_components = csg.connected_components(adj, directed=False,
                                                      return_labels=False)
    return True if n_connected_components == 1 else False


def neighbors(adj, area):
    """
    Parameters
    ----------
    adj : :class:`scipy.sparse.csr_matrix`
        Adjacency matrix.
    area : int
        An area specified as index in the graph represented by the adjacency
        matrix `adj`. The integer must be in `{0, 1, ..., adj.shape[0]-1}`.

    Returns
    -------
    neighs : :class:`numpy.ndarray`
        The neighbors of `area`.

    Examples
    --------
    >>> import numpy as np
    >>> from scipy.sparse import csr_matrix
    >>> adjacency_matrix = csr_matrix(np.array([[0, 1, 1],
    ...                                         [1, 0, 0],
    ...                                         [1, 0, 0]]))
    >>> (neighbors(adjacency_matrix, 0) == np.array([1, 2])).all()
    True
    >>> (neighbors(adjacency_matrix, 1) == np.array([0])).all()
    True
    """
    return adj[area].nonzero()[1]


def sub_adj_matrix(adj, nodes, wo_nodes=None):
    """
    Parameters
    ----------
    adj : :class:`scipy.sparse.csr_matrix`
        Adjacency matrix.
    nodes : :class:`numpy.ndarray`
        1-dimensional array of nodes in the graph represented by the `adj`
        argument. The elements in `nodes` are (distinct) integers in
        `{0, 1, ..., nodes.shape[0]}`.
    wo_nodes : slice, int, array of ints, or None
        Nodes to neglect from the `nodes` argument.

    Returns
    -------
    sub_adj : :class:`scipy.sparse.csr_matrix`
        Adjacency matrix of the subgraph consisting of only the nodes in the
        `nodes` argument.

    Examples
    --------
    >>> import numpy as np
    >>> from scipy.sparse import csr_matrix
    >>> adjacency_matrix = csr_matrix(np.array([[0, 1, 1],
    ...                                         [1, 0, 0],
    ...                                         [1, 0, 0]]))
    >>> nodes = np.array([0, 2])
    >>> obtained = sub_adj_matrix(adjacency_matrix, nodes)
    >>> desired = np.array([[0, 1],
    ...                     [1, 0]])
    >>> (obtained.todense() == desired).all()
    True
    >>> nodes = np.array([1, 2])
    >>> obtained = sub_adj_matrix(adjacency_matrix, nodes)
    >>> desired = np.array([[0, 0],
    ...                     [0, 0]])
    >>> (obtained.todense() == desired).all()
    True
    >>> type(obtained) == csr_matrix
    True
    >>> # tests for `wo_nodes` argument
    >>> all_nodes = np.arange(adjacency_matrix.shape[0])
    >>> neglected_nodes = np.array([1])
    >>> obtained = sub_adj_matrix(adjacency_matrix, all_nodes, neglected_nodes)
    >>> desired = np.array([[0, 1],
    ...                     [1, 0]])
    >>> (obtained.todense() == desired).all()
    True
    """
    if wo_nodes is not None:
        mask = np.in1d(nodes, wo_nodes, invert=True)
        nodes = nodes[mask]
    nodes = nodes[:, None]
    return csr_matrix(adj[nodes, nodes.T])




import networkx as nx
from libpysal import weights

def compare_region_lists(actual, desired):
    """
    Parameters
    ----------
    actual : list
        Every element (of type ``set``) represents a region.
    desired : list
        Every element (of type ``set``) represents a region.

    Raises
    ------
    AssertionError
        If the two arguments don't represent the same clustering.
    """
    # check number of regions
    assert len(actual) == len(desired)
    # check equality of regions
    assert all(region in desired for region in actual)


def region_list_from_array(labels):
    """

    Parameters
    ----------
    labels : :class:`numpy.ndarray`
        One-dimensional array of the areas' region labels.

    Returns
    -------
    region_list : `list`
        Each element is a `set` of areas belonging to the same region. The
        list's elements (the sets) are sorted by their smallest entry.

    Examples
    --------
    >>> import numpy as np
    >>> labels = np.array([0, 0, 0,
    ...                    1, 1, 0,
    ...                    1, 1, 1])
    >>> desired = [{0, 1, 2, 5}, {3, 4, 6, 7, 8}]
    >>> all(r1 == r2
    ...     for r1, r2 in zip(region_list_from_array(labels), desired))
    True
    """
    region_idx = {}  # mapping from region-ID to index in the returned list
    current_region_idx = 0
    region_list = []
    for area, region in enumerate(labels):
        if region not in region_idx:
            region_idx[region] = current_region_idx
            current_region_idx += 1
            region_list.append(set())
        region_list[region_idx[region]].add(area)
    return region_list


def convert_from_geodataframe(gdf):
    """
    Convert a GeoDataFrame to other types representing the contiguity relation
    of the GeoDataFrame's areas.

    Parameters
    ----------
    gdf : GeoDataFrame

    Returns
    -------
    other_formats : tuple
        The 1st entry is a sparse adjacency matrix of type
        :class:`scipy.sparse.csr_matrix`.
        The 2nd entry is a networkx graph.
        The 3rd entry is a dict. Each key is an area and each value is an
        iterable of the key area's neighbors.
        The 4th entry is a PySAL W object.
    """
    w = weights.Rook.from_dataframe(gdf)
    graph = w.to_networkx()
    adj = nx.to_scipy_sparse_matrix(graph)
    neighbor_dict = w.neighbors
    return adj, graph, neighbor_dict, w


from sklearn.metrics import pairwise as skm
from scipy.sparse import csgraph as cg
from scipy.optimize import OptimizeWarning
from collections import namedtuple
from warnings import warn
import time
import numpy as np
import copy
deletion = namedtuple('deletion', ('in_node', 'out_node', 'score'))

class Spanning_Forest(object):
    def __init__(self,
                 dissimilarity=skm.manhattan_distances,
                 affinity=None,
                 reduction=np.sum,
                 center=np.mean):
        """
        Initialize the SKATER algorithm.

        dissimilarity : a callable distance metric
        affinity : an callable affinity metric between 0,1. 
                   Will be inverted to provide a 
                   dissimilarity metric.
        reduction: the reduction applied over all clusters
                   to provide the map score.
        center:    way to compute the center of each region in attribute space
        
        NOTE: Optimization occurs with respect to a *dissimilarity* metric, so the reduction should
              yield some kind of score where larger values are *less desirable* than smaller values. 
              Typically, this means we use addition. 
        """
        if affinity is not None:
            # invert the 0,1 affinity to 
            # to an unbounded positive dissimilarity
            metric = lambda x: -np.log(affinity(x))
        else:
            metric = dissimilarity
        self.metric = metric
        self.reduction = reduction
        self.center = center
    
    def __repr__(self):
        return "Minimum_Spanning_Tree_Pruning(metric = {}, reduction = {}, center = {})".format(self.metric, 
                                                                                                self.reduction, 
                                                                                                self.center)
    
    def fit(self, n_clusters, W, 
            data=None, quorum=-np.inf, trace=False, islands='increase',
            verbose=False):
        """
        n_clusters : int of clusters wanted
        W : pysal W object expressing the neighbor relationships between observations. 
            Should be symmetric and binary, so Queen/Rook, DistanceBand, or a symmetrized KNN.
        data: np.ndarray of (N,P) shape with N observations and P features
        quorum: floor on the size of regions.
        trace: bool denoting whether to store intermediate
               labelings as the tree gets pruned
        islands: string describing what to do with islands. 
                 If "ignore", will discover `n_clusters` regions, treating islands as their own regions.
                 If "increase", will discover `n_clusters` regions, treating islands as separate from n_clusters. 

        verbose: bool/int describing how much output to provide to the user, 
                 in terms of print statements or progressbars.
        
        NOTE: Optimization occurs with respect to a *dissimilarity* metric, so the problem *minimizes* 
              the map dissimilarity. So, lower scores are better. 
        """
        if trace:
            self._trace = []
        if data is None:
            attribute_kernel = np.ones((W.n,W.n))
            data = np.ones((W.n,1))
        else:
            attribute_kernel = self.metric(data)
        W.transform = 'b'
        W = W.sparse
        start = time.time()

        super_verbose = verbose > 1
        start_W = time.time()
        dissim = W.multiply(attribute_kernel)
        dissim.eliminate_zeros()
        end_W = time.time() - start_W
        
        if super_verbose:
            print('Computing Affinity Kernel took {:.2f}s'.format(end_W))

        tree_time = time.time()
        MSF = cg.minimum_spanning_tree(dissim)
        tree_time = time.time() - tree_time
        if super_verbose:
            print('Computing initial MST took {:.2f}s'.format(tree_time))
        
        initial_component_time = time.time()
        current_n_subtrees, current_labels = cg.connected_components(MSF, directed=False)
        initial_component_time = time.time() - initial_component_time

        if super_verbose:
            print('Computing connected components took {:.2f}s.'.format(initial_component_time)) 
        
        if current_n_subtrees > 1:
            island_warnings = ['Increasing `n_clusters` from {} to {} in order to account for islands.'\
                              .format(n_clusters, n_clusters+current_n_subtrees),
                              'Counting islands towards the remaining {} clusters.'\
                              .format(n_clusters - (current_n_subtrees))]
            ignoring_islands = int(islands.lower() == 'ignore')
            chosen_warning = island_warnings[ignoring_islands]
            warn("By default, the graph is disconnected! {}".format(chosen_warning),
                 OptimizeWarning, stacklevel=2)
            if not ignoring_islands:
                n_clusters += (current_n_subtrees)      
            _,island_populations = np.unique(current_labels, return_counts=True)
            if (island_populations < quorum).any():
                raise ValueError("Islands must be larger than the quorum. If not, drop the small islands and solve for"
                                 " clusters in the remaining field.")
        if trace:
            self._trace.append((current_labels, deletion(np.nan, np.nan, np.inf)))
        if super_verbose:
            print(self._trace[-1])
        while current_n_subtrees < n_clusters: # while we don't have enough regions
            best_deletion = self.find_cut(MSF, data, quorum=quorum, 
                                          labels=None, target_label=None, verbose=verbose)

            if np.isfinite(best_deletion.score): # if our search succeeds
                # accept the best move as *the* move
                if super_verbose:
                    print('making cut {}...'.format(best_deletion))
                MSF, current_n_subtrees, current_labels = self.make_cut(*best_deletion, MSF=MSF)
            else: # otherwise, it means the MSF admits no further cuts (no backtracking here)
                current_n_subtrees, current_labels = cg.connected_components(MSF, directed=False)
                warn("MSF contains no valid moves after finding {} subtrees."
                     "Decrease the size of your quorum to find the remaining {} subtrees."
                     .format(current_n_subtrees, n_clusters - current_n_subtrees),
                     OptimizeWarning, stacklevel=2)
                self.current_labels_ = current_labels
                self.minimum_spanning_forest_ = MSF
                self._elapsed_time = time.time() - start
                return self
            if trace:
                self._trace.append((current_labels,best_deletion))
                
        self.current_labels_ = current_labels
        self.minimum_spanning_forest_ = MSF
        self._elapsed_time = time.time() - start
        return self
    
    def score(self, data, labels=None, quorum=-np.inf):
        """
        This yields a score for the data, given the labels provided. If no labels are provided,
        and the object has been fit, then the labels discovered from the previous fit are used. 
        
        If a quorum is not passed, it is assumed to be irrelevant. 
        
        If a quorum is passed and the labels do not meet quorum, the score is inf. 
       
        data    :   (N,P) array of data on which to compute the score of the regions expressed in labels
        labels  :   (N,) array of labels expressing the classification of each observation into a region.
        quorum  :   int expressing the minimum size of regions. Can be -inf if there is no lower bound. 
                    Any region below quorum makes the score inf.
        
        NOTE: Optimization occurs with respect to a *dissimilarity* metric, so the problem *minimizes* 
              the map dissimilarity. So, lower scores are better. 
        """
        if labels is None:
            try:
                labels = self.current_labels_
            except AttributeError:
                raise ValueError('Labels not provided and MSF_Prune object has not been fit to data yet.')
        assert data.shape[0] == len(labels), "Length of label array ({}) does not match "\
                                             "length of data ({})! ".format(labels.shape[0], data.shape[0])
        _, subtree_quorums = np.unique(labels, return_counts=True)
        n_subtrees = len(subtree_quorums)
        if (subtree_quorums < quorum).any():
            return np.inf
        part_scores = [self.reduction(self.metric(X=data[labels==l],
                                                  Y=self.center(data[labels==l], 
                                                                axis=0).reshape(1,-1)
                                                 )
                                     ) for l in range(n_subtrees)]
        return self.reduction(part_scores).item()
    
    def find_cut(self, MSF, data=None, quorum=-np.inf, 
                 labels=None, target_label=None, make=False, verbose=False):
        """
        Find the best cut from the MSF.
        
        MSF: (N,N) scipy sparse matrix with zero elements removed. 
             Represents the adjacency matrix for the minimum spanning forest. 
             Constructed from sparse.csgraph.sparse_from_dense or using MSF.eliminate_zeros(). 
             You MUST remove zero entries for this to work, otherwise they are considered no-cost paths. 
        data: (N,p) attribute matrix. If not provided, replaced with (N,1) vector of ones. 
        quorum: int denoting the minimum number of elements in the region
        labels: (N,) flat vector of labels for each point. Represents the "cluster labels" 
                for disconnected components of the graph. 
        target_label: int from the labels array to subset the MSF. If passed along with `labels`, then a cut
                      will be found that is restricted to that subset of the MSF.
        make: bool, whether or not to modify the input MSF in order to make the best cut that was found. 
        verbose: bool/int, denoting how much output to provide to the user, in terms
                 of print statements or progressbars
                      
        Returns a namedtuple with in_node, out_node, and score. 
        """
        if data is None:
            data = np.ones(MSF.shape)            
        
        if (labels is None) != (target_label is None):
            raise ValueError("Both labels and target_label must be supplied! Only {} provided."\
                             .format(['labels', 'target_label'][int(target_label is None)]))
        if verbose:
            try:
                from tqdm import tqdm
            except ImportError:
                def tqdm(noop, desc=''):
                    return noop
        else:
            def tqdm(noop, desc=''):
                return noop
        zero_in = (labels is not None) and (target_label is not None)
        current_n_subtrees, current_labels = cg.connected_components(MSF, directed=False)
        best_deletion = deletion(np.nan, np.nan, np.inf)
        for in_node, out_node in tqdm(np.vstack(MSF.nonzero()).T, desc='finding cut...'): # iterate over MSF edges
            if zero_in:
                if labels[in_node] != target_label:
                    continue

            local_MSF = copy.deepcopy(MSF)
                
            # delete a candidate edge
            local_MSF[in_node, out_node] = 0
            local_MSF.eliminate_zeros()
             
            # get the connected components
            local_n_subtrees, local_labels = cg.connected_components(local_MSF, directed=False)
                
            if local_n_subtrees <= current_n_subtrees:
                raise Exception('Malformed MSF!')
                
            # compute the score of these components
            score = self.score(data, labels=local_labels, quorum=quorum)
                
            # if the score is lower than the best score and quorum is met
            if score < best_deletion.score:
                best_deletion = deletion(in_node, out_node, score)
        if make:
            return self.make_cut(*best_deletion, MSF=MSF)
        return best_deletion
    
    def make_cut(self, in_node, out_node, score, MSF=None):
        """
        make a cut on the MSF inplace, provided the in_node, out_node, MSF, and score. 
        in_node: int, ID of the source node for the edge to be cut
        out_node: int, ID of the destination node for the edge to be cut
        score: float, the value of the score being cut. if the score is infinite, the cut is not made. 
        MSF: the spanning forest to use when making the cut. If not provided,
             uses the defualt tree in self.minimum_spanning_forest_
        """
        if MSF is None:
            MSF = self.minimum_spanning_forest_
        if np.isfinite(score):
            MSF[in_node, out_node] = 0
            MSF.eliminate_zeros()
            return (MSF, *cg.connected_components(MSF, directed=False))
        raise OptimizeWarning('Score of the ({},{}) cut is inf, the quorum is likely not met!')
        
    
if __name__ == "__main__":
    
    from libpysal import examples, weights
    import geopandas as gpd
    
    df = gpd.read_file(examples.get_path('NAT.shp'))
    data = df[df.filter(like='90').columns.tolist()\
           + df.filter(like='89').columns.tolist()].values
    data_c = (data - data.mean(axis=0)) / data.std(axis=0)
    W = weights.Queen.from_dataframe(df)
    result = Spanning_Forest().fit(10, W, data_c, quorum=100)
    
    will_fail = Spanning_Forest().fit(10,W,data_c, quorum=500)



import abc
from collections import deque
import math
import random

import numpy as np
import networkx as nx

from region.csgraph_utils import sub_adj_matrix, neighbors, is_connected
from region.objective_function import ObjectiveFunctionPairwise
from region.p_regions.azp_util import AllowMoveStrategy, \
                                            AllowMoveAZP,\
                                            AllowMoveAZPSimulatedAnnealing
from region.util import array_from_df_col, array_from_dict_values, \
    assert_feasible, boolean_assert_feasible, copy_func, count, generate_initial_sol, \
    make_move, Move, pop_randomly_from, random_element_from,\
    scipy_sparse_matrix_from_w, separate_components, w_from_gdf,\
    array_from_graph_or_dict, scipy_sparse_matrix_from_dict


class AZP:
    """
    Class offering the implementation of the AZP algorithm (see [OR1995]_).

    Attributes
    ----------
    labels_ : :class:`numpy.ndarray`
        Each element is a region label specifying to which region the
        corresponding area was assigned to by the last run of a fit-method.
    """

    def __init__(self, allow_move_strategy=None, random_state=None):
        """
        Parameters
        ----------
        allow_move_strategy : None or :class:`AllowMoveStrategy`, default: None
            If None, then the AZP algorithm in [OR1995]_ is chosen.
            For a different behavior for allowing moves an AllowMoveStrategy
            instance can be passed as argument.
        random_state : None, int, str, bytes, or bytearray, default: None
            Random seed.
        """
        self.n_regions = None
        self.labels_ = None
        self.random_state = random_state
        random.seed(self.random_state)

        if isinstance(allow_move_strategy, AllowMoveStrategy):
            self.allow_move_strategy = allow_move_strategy
        elif allow_move_strategy is None:
            self.allow_move_strategy = AllowMoveAZP()
        else:
            raise ValueError("The allow_move_strategy argument must be either "
                             "None, or an instance of AllowMoveStrategy.")

        self.objective_func = None

    def fit_from_scipy_sparse_matrix(
            self,
            adj,
            attr,
            n_regions,
            initial_labels=None,
            objective_func=ObjectiveFunctionPairwise()):
        """
        Perform the AZP algorithm as described in [OR1995]_.

        The resulting region labels are assigned to the instance's
        :attr:`labels_` attribute.

        Parameters
        ----------
        adj : :class:`scipy.sparse.csr_matrix`
            Adjacency matrix representing the contiguity relation.
        attr : :class:`numpy.ndarray`
            Array (number of areas x number of attributes) of areas' attributes
            relevant to clustering.
        n_regions : `int`
            Number of desired regions.
        initial_labels : :class:`numpy.ndarray` or None, default: None
            One-dimensional array of labels at the beginning of the algorithm.
            If None, then a random initial clustering will be generated.
        objective_func : :class:`region.objective_function.ObjectiveFunction`, default: ObjectiveFunctionPairwise()
            The objective function to use.
        """
        if attr.ndim == 1:
            attr = attr.reshape(adj.shape[0], -1)
        self.allow_move_strategy.attr_all = attr
        self.objective_func = objective_func
        # step 1
        if initial_labels is not None:
            assert_feasible(initial_labels, adj, n_regions)
            initial_labels_gen = separate_components(adj, initial_labels)
        else:
            initial_labels_gen = generate_initial_sol(adj, n_regions)
        labels = -np.ones(adj.shape[0])
        for labels_comp in initial_labels_gen:
            comp_idx = np.where(labels_comp != -1)[0]
            adj_comp = sub_adj_matrix(adj, comp_idx)
            labels_comp = labels_comp[comp_idx]
            attr_comp = attr[comp_idx]
            self.allow_move_strategy.start_new_component(
                labels_comp, attr_comp, self.objective_func, comp_idx)

            labels_comp = self._azp_connected_component(
                adj_comp, labels_comp, attr_comp)
            labels[comp_idx] = labels_comp

        self.n_regions = n_regions
        self.labels_ = labels

    fit = copy_func(fit_from_scipy_sparse_matrix)
    fit.__doc__ = "Alias for :meth:`fit_from_scipy_sparse_matrix`.\n\n" \
                  + fit_from_scipy_sparse_matrix.__doc__

    def fit_from_w(self,
                   w,
                   attr,
                   n_regions,
                   initial_labels=None,
                   objective_func=ObjectiveFunctionPairwise()):
        """
        Alternative API for :meth:`fit_from_scipy_sparse_matrix`.

        Parameters
        ----------
        w : :class:`libpysal.weights.weights.W`
            W object representing the contiguity relation.
        attr : :class:`numpy.ndarray`
            Refer to the corresponding argument in
            :meth:`fit_from_scipy_sparse_matrix`.
        n_regions : `int`
            Refer to the corresponding argument in
            :meth:`fit_from_scipy_sparse_matrix`.
        initial_labels : :class:`numpy.ndarray` or None, default: None
            Refer to the corresponding argument in
            :meth:`fit_from_scipy_sparse_matrix`.
        objective_func : :class:`region.ObjectiveFunction`, default: ObjectiveFunctionPairwise()
            Refer to the corresponding argument in
            :meth:`fit_from_scipy_sparse_matrix`.
        """
        adj = scipy_sparse_matrix_from_w(w)
        self.fit_from_scipy_sparse_matrix(
            adj,
            attr,
            n_regions,
            initial_labels,
            objective_func=objective_func)

    def fit_from_networkx(self,
                          graph,
                          attr,
                          n_regions,
                          initial_labels=None,
                          objective_func=ObjectiveFunctionPairwise()):
        """
        Alternative API for :meth:`fit_from_scipy_sparse_matrix`.

        Parameters
        ----------
        graph : `networkx.Graph`
            Graph representing the contiguity relation.
        attr : str, list or dict
            If the clustering criteria are present in the networkx.Graph
            `graph` as node attributes, then they can be specified as a string
            (for one criterion) or as a list of strings (for multiple
            criteria).
            Alternatively, a dict can be used with each key being a node of the
            networkx.Graph `graph` and each value being the corresponding
            clustering criterion (a scalar (e.g. `float` or `int`) or a
            :class:`numpy.ndarray`).
            If there are no clustering criteria present in the networkx.Graph
            `graph` as node attributes, then a dictionary must be used for this
            argument. Refer to the corresponding argument in
            :meth:`fit_from_dict` for more details about the expected dict.
        n_regions : `int`
            Refer to the corresponding argument in
            :meth:`fit_from_scipy_sparse_matrix`.
        initial_labels : str or dict or None, default: None
            If str, then the string names the graph's attribute holding the
            information about the initial clustering.
            If dict, then each key is a node and each value is the region the
            key area is assigned to at the beginning of the algorithm.
            If None, then a random initial clustering will be generated.
        objective_func : :class:`region.ObjectiveFunction`, default: ObjectiveFunctionPairwise()
            Refer to the corresponding argument in
            :meth:`fit_from_scipy_sparse_matrix`.
        """
        adj = nx.to_scipy_sparse_matrix(graph)
        attr = array_from_graph_or_dict(graph, attr)
        if initial_labels is not None:
            initial_labels = array_from_graph_or_dict(graph, initial_labels)
        self.fit_from_scipy_sparse_matrix(
            adj,
            attr,
            n_regions,
            initial_labels,
            objective_func=objective_func)

    def fit_from_geodataframe(self,
                              gdf,
                              attr,
                              n_regions,
                              contiguity="rook",
                              initial_labels=None,
                              objective_func=ObjectiveFunctionPairwise()):
        """
        Alternative API for :meth:`fit_from_scipy_sparse_matrix`.

        Parameters
        ----------
        gdf : :class:`geopandas.GeoDataFrame`

        attr : `str` or `list`
            The clustering-relevant attributes (columns of the GeoDataFrame
            `gdf`) are specified as string (for one column) or list of strings
            (for multiple columns).
        n_regions : `int`
            Refer to the corresponding argument in
            :meth:`fit_from_scipy_sparse_matrix`.
        contiguity : {"rook", "queen"}, default: "rook"
            Defines the contiguity relationship between areas. Possible
            contiguity definitions are:

            * "rook" - Rook contiguity.
            * "queen" - Queen contiguity.

        initial_labels : :class:`numpy.ndarray` or None, default: None
            Refer to the corresponding argument in
            :meth:`fit_from_scipy_sparse_matrix`.
        objective_func : :class:`region.ObjectiveFunction`, default: ObjectiveFunctionPairwise()
            Refer to the corresponding argument in
            :meth:`fit_from_scipy_sparse_matrix`.
        """
        w = w_from_gdf(gdf, contiguity)
        attr = array_from_df_col(gdf, attr)
        self.fit_from_w(
            w, attr, n_regions, initial_labels, objective_func=objective_func)

    def fit_from_dict(self,
                      neighbor_dict,
                      attr,
                      n_regions,
                      initial_labels=None,
                      objective_func=ObjectiveFunctionPairwise()):
        """
        Alternative API for :meth:`fit_from_scipy_sparse_matrix`.

        Parameters
        ----------
        neighbor_dict : `dict`
            Each key is an area and each value is an iterable of the key area's
            neighbors.
        attr : `dict`
            Each key is an area and each value is the corresponding
            clustering-relevant attribute.
        n_regions : `int`
            Refer to the corresponding argument in
            :meth:`fit_from_scipy_sparse_matrix`.
        initial_labels : `dict` or None, default: None
            Each key represents an area. Each value represents the region, the
            corresponding area is assigned to at the beginning of the
            algorithm.
            If None, then a random initial clustering will be generated.
        objective_func : :class:`region.ObjectiveFunction`, default: ObjectiveFunctionPairwise()
            Refer to the corresponding argument in
            :meth:`fit_from_scipy_sparse_matrix`.
        """
        sorted_areas = sorted(neighbor_dict)

        adj = scipy_sparse_matrix_from_dict(neighbor_dict)
        attr_arr = array_from_dict_values(attr, sorted_areas)

        if initial_labels is not None:
            initial_labels = array_from_dict_values(
                initial_labels, sorted_areas, flat_output=True, dtype=np.int32)
        self.fit_from_scipy_sparse_matrix(
            adj,
            attr_arr,
            n_regions,
            initial_labels,
            objective_func=objective_func)

    def _azp_connected_component(self, adj, initial_clustering, attr):
        """
        Implementation of the AZP algorithm for a spatially connected set of
        areas (i.e. for every area there is a path to every other area).

        Parameters
        ----------
        adj : :class:`scipy.sparse.csr_matrix`
            Adjacency matrix representing the contiguity relation. The matrix'
            shape is (N, N) where N denotes the number of areas in the
            currently considered connected component.
        initial_clustering : :class:`numpy.ndarray`
            Array of labels. Shape: (N,) where N denotes the number of areas in
            the currently considered connected component.
        attr : :class:`numpy.ndarray`
            Array of labels. Shape: (N, M) where N denotes the number of areas
            in the currently considered connected component and M denotes the
            number of attributes per area.

        Returns
        -------
        labels : :class:`numpy.ndarray`
            One-dimensional array of region labels after the AZP algorithm has
            been performed. Only region labels of the currently considered
            connected component are returned.
        """
        # if there is only one region in the initial solution, just return it.
        distinct_regions = list(np.unique(initial_clustering))
        if len(distinct_regions) == 1:
            return initial_clustering
        distinct_regions_copy = distinct_regions.copy()

        #  step 2: make a list of the M regions
        labels = initial_clustering

        obj_val_start = float("inf")
        obj_val_end = self.allow_move_strategy.objective_val

        region_neighbors = {}
        for region in distinct_regions:
            region_areas = set(np.where(labels == region)[0])
            neighs = set()
            for area in region_areas:
                neighs.update(neighbors(adj, area))
            region_neighbors[region] = neighs.difference(region_areas)
        del neighs

        # step 7: Repeat until no further improving moves are made
        while obj_val_end < obj_val_start:  # improvement
            obj_val_start = float(obj_val_end)
            distinct_regions = distinct_regions_copy.copy()
            # step 6: when the list for region K is exhausted return to step 3
            # and select another region and repeat steps 4-6

            while distinct_regions:
                # step 3: select & remove any region K at random from this list
                recipient = pop_randomly_from(distinct_regions)
                while True:
                    # step 4: identify a set of zones bordering on members of
                    # region K that could be moved into region K without
                    # destroying the internal contiguity of the donor region(s)

                    candidates = []
                    for neigh in region_neighbors[recipient]:
                        neigh_region = labels[neigh]
                        sub_adj = sub_adj_matrix(
                            adj,
                            np.where(labels == neigh_region)[0],
                            wo_nodes=neigh)
                        if is_connected(sub_adj):
                            # if area is alone in its region, it must stay
                            if count(labels, neigh_region) > 1:
                                candidates.append(neigh)
                    # step 5: randomly select zones from this list until either
                    # there is a local improvement in the current value of the
                    # objective function or a move that is equivalently as good
                    # as the current best. Then make the move, update the list
                    # of candidate zones, and return to step 4 or else repeat
                    # step 5 until the list is exhausted.
                    while candidates:
                        cand = pop_randomly_from(candidates)
                        if self.allow_move_strategy(cand, recipient, labels):
                            donor = labels[cand]

                            make_move(cand, recipient, labels)

                            region_neighbors[donor].add(cand)
                            region_neighbors[recipient].discard(cand)

                            neighs_of_cand = neighbors(adj, cand)

                            recipient_region_areas = set(
                                np.where(labels == recipient)[0])
                            region_neighbors[recipient].update(neighs_of_cand)
                            region_neighbors[recipient].difference_update(
                                recipient_region_areas)

                            donor_region_areas = set(
                                np.where(labels == donor)[0])
                            not_donor_neighs_anymore = set(
                                area for area in neighs_of_cand if not any(
                                    a in donor_region_areas
                                    for a in neighbors(adj, area)))
                            region_neighbors[donor].difference_update(
                                not_donor_neighs_anymore)
                            break
                    else:
                        break

            obj_val_end = float(self.allow_move_strategy.objective_val)
        return labels


class AZPSimulatedAnnealing:
    """
    Class offering the implementation of the AZP-SA algorithm (see [OR1995]_).

    Attributes
    ----------
    labels_ : :class:`numpy.ndarray`
        Each element is a region label specifying to which region the
        corresponding area was assigned to by the last run of a fit-method.
    """

    def __init__(self,
                 init_temperature=None,
                 max_iterations=float("inf"),
                 sa_moves_term=10,
                 nonmoving_steps_before_stop=3,
                 repetitions_before_termination=5,
                 random_state=None):
        """
        Parameters
        ----------
        init_temperature : float
            The initial temperature used in the simulated annealing algorithm.
        max_iterations : int or {float("inf")}, default: float("inf")
            Termination condition for step b: Terminate if the AZP algorithm
            has run for `max_iterations` times.
        sa_moves_term : int, default: 0
            Termination condition for step b: Count the SA-moves made by the
            repeated runs of the AZP (modified in step 5) and terminate after
            the AZP run that made the cumulative number of SA-moves reach or
            exceed `sa_moves_term`.
        nonmoving_steps_before_stop : int, default: 3
            Termination condition: Repeat steps b and c until no further moves
            occur for `nonmoving_steps_before_stop` times.
        repetitions_before_termination : int, default: 5
            Termination condition not present in [OR1995]_: Terminate if the
            AZP runs returned a given solution for
            `repetitions_before_termination` times.
        random_state : None, int, str, bytes, or bytearray
            Random seed.
        """
        self.allow_move_strategy = None
        self.azp = None

        if init_temperature is not None:
            self.init_temperature = init_temperature
        else:
            raise NotImplementedError("TODO")  # todo

        self.maxit = max_iterations
        self.sa_moves_term = sa_moves_term

        self.sa_moves_term_reached = False
        self.move_made = False
        self.nonmoving_steps_before_stop = nonmoving_steps_before_stop

        self.visited = []
        self.reps_before_termination = repetitions_before_termination

        self.random_state = random_state

        self.n_regions = None

    def fit_from_geodataframe(self,
                              gdf,
                              attr,
                              n_regions,
                              contiguity="rook",
                              initial_labels=None,
                              cooling_factor=0.85,
                              objective_func=ObjectiveFunctionPairwise()):
        """
        Parameters
        ----------
        gdf : :class:`geopandas.GeoDataFrame`
            Refer to the corresponding argument in
            :meth:`AZP.fit_from_geodataframe`.
        attr : `str` or `list`
            Refer to the corresponding argument in
            :meth:`AZP.fit_from_geodataframe`.
        n_regions : `int`
            Refer to the corresponding argument in
            :meth:`AZP.fit_from_geodataframe`.
        contiguity : `str`
            Refer to the corresponding argument in
            :meth:`AZP.fit_from_geodataframe`.
        initial_labels : :class:`numpy.ndarray` or None, default: None
            Refer to the corresponding argument in
            :meth:`AZP.fit_from_geodataframe`.
        cooling_factor : float, default: 0.85
            Refer to the corresponding argument in
            :meth:`fit_from_scipy_sparse_matrix`.
        objective_func : :class:`region.ObjectiveFunction`, default: ObjectiveFunctionPairwise()
            Refer to the corresponding argument in
            :meth:`fit_from_scipy_sparse_matrix`.
        """
        w = w_from_gdf(gdf, contiguity)
        attr = array_from_df_col(gdf, attr)
        self.fit_from_w(
            w,
            attr,
            n_regions,
            initial_labels,
            cooling_factor=cooling_factor,
            objective_func=objective_func)

    def fit_from_dict(self,
                      neighbor_dict,
                      attr,
                      n_regions,
                      initial_labels=None,
                      cooling_factor=0.85,
                      objective_func=ObjectiveFunctionPairwise()):
        """
        Parameters
        ----------
        neighbor_dict : `dict`
            Refer to the corresponding argument in :meth:`AZP.fit_from_dict`.
        attr : `dict`
            Refer to the corresponding argument in :meth:`AZP.fit_from_dict`.
        n_regions : `int`
            Refer to the corresponding argument in
            :meth:`fit_from_scipy_sparse_matrix`.
        initial_labels : `dict` or None, default: None
            Refer to the corresponding argument in :meth:`AZP.fit_from_dict`.
        cooling_factor : float, default: 0.85
            Refer to the corresponding argument in
            :meth:`fit_from_scipy_sparse_matrix`.
        objective_func : :class:`region.ObjectiveFunction`, default: ObjectiveFunctionPairwise()
            Refer to the corresponding argument in
            :meth:`fit_from_scipy_sparse_matrix`.
        """
        sorted_areas = sorted(neighbor_dict)
        adj = scipy_sparse_matrix_from_dict(neighbor_dict)
        attr_arr = array_from_dict_values(attr, sorted_areas)

        if initial_labels is not None:
            initial_labels = array_from_dict_values(
                initial_labels, sorted_areas, flat_output=True, dtype=np.int32)
        self.fit_from_scipy_sparse_matrix(
            adj,
            attr_arr,
            n_regions,
            initial_labels=initial_labels,
            cooling_factor=cooling_factor,
            objective_func=objective_func)

    def fit_from_networkx(self,
                          graph,
                          attr,
                          n_regions,
                          initial_labels=None,
                          cooling_factor=0.85,
                          objective_func=ObjectiveFunctionPairwise()):
        """
        Parameters
        ----------
        graph : `networkx.Graph`
            Refer to the corresponding argument in
            :meth:`AZP.fit_from_networkx`.
        attr : str, list or dict
            Refer to the corresponding argument in
            :meth:`AZP.fit_from_networkx`.
        n_regions : `int`
            Refer to the corresponding argument in
            :meth:`fit_from_scipy_sparse_matrix`.
        initial_labels : str or dict or None, default: None
            Refer to the corresponding argument in
            :meth:`AZP.fit_from_networkx`.
        cooling_factor : float, default: 0.85
            Refer to the corresponding argument in
            :meth:`fit_from_scipy_sparse_matrix`.
        objective_func : :class:`region.ObjectiveFunction`, default: ObjectiveFunctionPairwise()
            Refer to the corresponding argument in
            :meth:`fit_from_scipy_sparse_matrix`.
        """
        adj = nx.to_scipy_sparse_matrix(graph)
        attr = array_from_graph_or_dict(graph, attr)
        if initial_labels is not None:
            initial_labels = array_from_graph_or_dict(graph, initial_labels)
        self.fit_from_scipy_sparse_matrix(
            adj,
            attr,
            n_regions,
            initial_labels,
            cooling_factor=cooling_factor,
            objective_func=objective_func)

    def fit_from_scipy_sparse_matrix(
            self,
            adj,
            attr,
            n_regions,
            initial_labels=None,
            cooling_factor=0.85,
            objective_func=ObjectiveFunctionPairwise()):
        """
        Parameters
        ----------
        adj : :class:`scipy.sparse.csr_matrix`
            Refer to the corresponding argument in
            :meth:`AZP.fit_from_scipy_sparse_matrix`.
        attr : :class:`numpy.ndarray`
            Refer to the corresponding argument in
            :meth:`AZP.fit_from_scipy_sparse_matrix`.
        n_regions : `int`
            Refer to the corresponding argument in
            :meth:`AZP.fit_from_scipy_sparse_matrix`.
        initial_labels : :class:`numpy.ndarray` or None, default: None
            Refer to the corresponding argument in
            :meth:`AZP.fit_from_scipy_sparse_matrix`.
        cooling_factor : float, default: 0.85
            Float :math:`\\in (0, 1)` specifying the cooling factor for the
            simulated annealing.
        objective_func : :class:`region.ObjectiveFunction`, default: ObjectiveFunctionPairwise()
            Refer to the corresponding argument in
            :meth:`fit_from_scipy_sparse_matrix`.
        """
        if not (0 < cooling_factor < 1):
            raise ValueError("The cooling_factor argument must be greater "
                             "than 0 and less than 1")
        if attr.ndim == 1:
            attr = attr.reshape(adj.shape[0], -1)
        self.allow_move_strategy = AllowMoveAZPSimulatedAnnealing(
            init_temperature=self.init_temperature,
            sa_moves_term=self.sa_moves_term)
        self.allow_move_strategy.register_sa_moves_term(self.sa_moves_alert)
        self.allow_move_strategy.register_move_made(self.move_made_alert)

        self.azp = AZP(
            allow_move_strategy=self.allow_move_strategy,
            random_state=self.random_state)
        # step a
        t = self.init_temperature
        nonmoving_steps = 0
        # step d: repeat step b and c
        while nonmoving_steps < self.nonmoving_steps_before_stop:
            it = 0
            self.sa_moves_term_reached = False
            self.allow_move_strategy.reset()
            # step b
            while it < self.maxit and not self.sa_moves_term_reached:
                it += 1
                old_sol = initial_labels
                self.azp.fit_from_scipy_sparse_matrix(
                    adj, attr, n_regions, initial_labels, objective_func)
                initial_labels = self.azp.labels_

                if old_sol is not None:

                    if (old_sol == initial_labels).all():
                        break
            # added termination condition (not in Openshaw & Rao (1995))
            if self.visited.count(tuple(initial_labels)) \
                    >= self.reps_before_termination:
                break
            self.visited.append(tuple(initial_labels))
            # step c
            t *= cooling_factor
            self.allow_move_strategy.update_temperature(t)

            if self.move_made:
                self.move_made = False
            else:
                nonmoving_steps += 1
        self.labels_ = initial_labels

    def fit_from_w(self,
                   w,
                   attr,
                   n_regions,
                   initial_labels=None,
                   cooling_factor=0.85,
                   objective_func=ObjectiveFunctionPairwise()):
        """
        Parameters
        ----------
        w : :class:`libpysal.weights.weights.W`
            Refer to the corresponding argument in :meth:`AZP.fit_from_w`.
        attr : :class:`numpy.ndarray`
            Refer to the corresponding argument in
            :meth:`fit_from_scipy_sparse_matrix`.
        n_regions : `int`
            Refer to the corresponding argument in
            :meth:`fit_from_scipy_sparse_matrix`.
        initial_labels : :class:`numpy.ndarray` or None, default: None
            Refer to the corresponding argument in
            :meth:`fit_from_scipy_sparse_matrix`.
        cooling_factor : float, default: 0.85
            Refer to the corresponding argument in
            :meth:`fit_from_scipy_sparse_matrix`.
        objective_func : :class:`region.ObjectiveFunction`, default: ObjectiveFunctionPairwise()
            Refer to the corresponding argument in
            :meth:`fit_from_scipy_sparse_matrix`.
        """
        adj = scipy_sparse_matrix_from_w(w)
        self.fit_from_scipy_sparse_matrix(
            adj,
            attr,
            n_regions,
            initial_labels,
            cooling_factor=cooling_factor,
            objective_func=objective_func)

    def sa_moves_alert(self):
        self.sa_moves_term_reached = True

    def move_made_alert(self):
        self.move_made = True


class AZPTabu(AZP, abc.ABC):
    """
    Superclass for tabu variants of the AZP.
    """

    def _make_move(self, area, new_region, labels, adj):
        old_region = labels[area]
        make_move(area, new_region, labels)
        if not boolean_assert_feasible(labels, adj): # If the move breaks the contiguity, revert it!
            make_move(area, old_region, labels) # Revert Move!
            reverse_move = Move(area, new_region, old_region)
            self.tabu.append(reverse_move)
            return False
        else:
            # step 5: Tabu the reverse move for R iterations.
            reverse_move = Move(area, new_region, old_region)
            self.tabu.append(reverse_move)
            return True

    def reset_tabu(self, tabu_len=None):
        tabu_len = self.tabu.maxlen if tabu_len is None else tabu_len
        self.tabu = deque([], tabu_len)


class AZPBasicTabu(AZPTabu):
    """
    Implementation of the AZP with basic tabu (refer to [OR1995]_).

    Attributes
    ----------
    labels_ : :class:`numpy.ndarray`
        Each element is a region label specifying to which region the
        corresponding area was assigned to by the last run of a fit-method.
    """

    def __init__(self,
                 tabu_length=None,
                 repetitions_before_termination=5,
                 random_state=None):
        """
        Parameters
        ----------
        tabu_length : numbers.Integral
            The size of the tabu list.
        repetitions_before_termination : numbers.integral
            This argument specifies a termination condition. If a solution has
            been visited for `repetitions_before_termination` times, the
            clustering function will terminate.
        random_state : None, int, str, bytes, or bytearray, default: None
            Refer to the corresponding argument in
            :meth:`AZP.__init__`.
        """
        self.tabu = deque([], tabu_length)
        self.visited = []
        self.reps_before_termination = repetitions_before_termination
        super().__init__(random_state=random_state)

    def _azp_connected_component(self, adj, initial_clustering, attr):
        """
        Implementation of the basic tabu version of the AZP algorithm (refer
        to [OR1995]_) for a spatially connected set of areas (i.e. for every
        area there is a path to every other area).

        Parameters
        ----------
        adj : :class:`scipy.sparse.csr_matrix`
            Refer to the corresponding argument in
            :meth:`AZP._azp_connected_component`.
        initial_clustering : :class:`numpy.ndarray`
            Refer to the corresponding argument in
            :meth:`AZP._azp_connected_component`.
        attr : :class:`numpy.ndarray`
            Refer to the corresponding argument in
            :meth:`AZP._azp_connected_component`.

        Returns
        -------
        labels : :class:`numpy.ndarray`
            Refer to the return value in :meth:`AZP._azp_connected_component`.
        """
        self.reset_tabu()
        # if there is only one region in the initial solution, just return it.
        distinct_regions = list(np.unique(initial_clustering))
        if len(distinct_regions) == 1:
            return initial_clustering

        #  step 2: make a list of the M regions
        labels = initial_clustering

        visited = []
        stop = False
        while True:
            # added termination condition (not in Openshaw & Rao (1995))
            label_tup = tuple(labels)
            if visited.count(label_tup) >= self.reps_before_termination:
                stop = True

            visited.append(label_tup)

            # step 1 Find the global best move that is not prohibited or tabu.
            # find possible moves (globally)
            best_move = None
            best_objval_diff = float("inf")
            for area in range(labels.shape[0]):
                old_region = labels[area]
                sub_adj = sub_adj_matrix(
                    adj, np.where(labels == old_region)[0], wo_nodes=area)
                # moving the area must not destroy spatial contiguity in donor
                # region and if area is alone in its region, it must stay:
                if is_connected(sub_adj) and count(labels, old_region) > 1:
                    for neigh in neighbors(adj, area):
                        new_region = labels[neigh]
                        if new_region != old_region:
                            possible_move = Move(area, old_region, new_region)
                            if possible_move not in self.tabu:
                                objval_diff = self.objective_func.update(
                                    possible_move.area,
                                    possible_move.new_region, labels, attr)
                                if objval_diff < best_objval_diff:
                                    best_move = possible_move
                                    best_objval_diff = objval_diff
            # step 2: Make this move if it is an improvement or equivalet in
            # value.
            if best_move is not None and best_objval_diff <= 0 and self.allow_move_strategy(best_move.area, best_move.new_region, labels):
                self._make_move(best_move.area, best_move.new_region, labels, adj)
            else:
                # step 3: if no improving move can be made, then see if a tabu
                # move can be made which improves on the current local best
                # (termed an aspiration move)
                improving_tabus = [
                    move for move in self.tabu
                    if labels[move.area] == move.old_region
                    and self.objective_func.update(move.area, move.new_region,
                                                   labels, attr) < 0
                ]
                if improving_tabus:
                    aspiration_move = random_element_from(improving_tabus)
                    
                    if self.allow_move_strategy(aspiration_move.area, aspiration_move.new_region, labels):
                    
                        self._make_move(aspiration_move.area, aspiration_move.new_region, labels, adj)
                        
                if stop:
                        break
                else:
                    # step 4: If there is no improving move and no aspirational
                    # move, then make the best move even if it is nonimproving
                    # (that is, results in a worse value of the objective
                    # function).
                    if stop:
                        break
                    if best_move is not None and self.allow_move_strategy(best_move.area, best_move.new_region, labels):
                        self._make_move(best_move.area, best_move.new_region, labels, adj)
        return labels


class AZPReactiveTabu(AZPTabu):
    """
    Implementation of the AZP with reactive tabu (refer to [OR1995]_).

    Attributes
    ----------
    labels_ : :class:`numpy.ndarray`
        Each element is a region label specifying to which region the
        corresponding area was assigned to by the last run of a fit-method.
    """

    def __init__(self, max_iterations, k1, k2, random_state=None):
        """
        Parameters
        ----------
        max_iterations : int
           Termination condition: The algorithm terminates after steps 3-11
           (see [OR1995]_) are repeated for `max_max_iterations` times.
        k1 : int
            Defining a necessary condition for jumping from step 7 to step 11
            in the algorithm (see [OR1995]_). Such a jump requires (besides the
            condition involving `k2`) a solution to be visited more than `k1`
            times.
        k2 : int
            Defining a necessary condition for jumping from step 7 to step 11
            in the algorithm (see [OR1995]_). Such a jump requires (besides the
            condition involving `k1`) a cycle of solutions to be found at least
            `k2` times.
        random_state : None, int, str, bytes, or bytearray, default: None
            Refer to the corresponding argument in
            :meth:`AZP.__init__`.
        """
        self.tabu = deque([], maxlen=1)
        super().__init__(random_state=random_state)
        self.avg_it_until_rep = 1
        self.rep_counter = 1
        if max_iterations <= 0:
            raise ValueError("The `max_iterations` argument must be > 0.")
        self.maxit = max_iterations
        self.visited = []
        self.k1 = k1
        self.k2 = k2

    def _azp_connected_component(self, adj, initial_labels, attr):
        """
        Implementation of the reactive tabu version of the AZP algorithm (refer
        to [OR1995]_) for a spatially connected set of areas (i.e. for every
        area there is a path to every other area).

        Parameters
        ----------
        adj : :class:`scipy.sparse.csr_matrix`
            Refer to the corresponding argument in
            :meth:`AZP._azp_connected_component`.
        initial_labels : :class:`numpy.ndarray`
            Refer to the corresponding argument in
            :meth:`AZP._azp_connected_component`.
        attr : :class:`numpy.ndarray`
            Refer to the corresponding argument in
            :meth:`AZP._azp_connected_component`.

        Returns
        -------
        labels : :class:`numpy.ndarray`
            Refer to the return value in :meth:`AZP._azp_connected_component`.
        """
        self.reset_tabu(1)
        # if there is only one region in the initial solution, just return it.
        distinct_regions = list(np.unique(initial_labels))
        if len(distinct_regions) == 1:
            return initial_labels

        #  step 2: make a list of the M regions
        labels = initial_labels

        it_since_tabu_len_changed = 0
        obj_val_start = float("inf")
        # step 12: Repeat steps 3-11 until either no further improvements are
        # made or a maximum number of iterations are exceeded.
        for it in range(self.maxit):
            obj_val_end = self.objective_func(labels, attr)
            if not obj_val_end < obj_val_start:
                break  # step 12
            obj_val_start = obj_val_end

            it_since_tabu_len_changed += 1
            # step 3: Define the list of all possible moves that are not tabu
            # and retain regional connectivity.
            possible_moves = []
            for area in range(labels.shape[0]):
                old_region = labels[area]
                sub_adj = sub_adj_matrix(
                    adj, np.where(labels == old_region)[0], wo_nodes=area)
                # moving the area must not destroy spatial contiguity in donor
                # region and if area is alone in its region, it must stay:
                if is_connected(sub_adj) and count(labels, old_region) > 1:
                    for neigh in neighbors(adj, area):
                        new_region = labels[neigh]
                        if new_region != old_region:
                            possible_move = Move(area, old_region, new_region)
                            if possible_move not in self.tabu:
                                possible_moves.append(possible_move)
            # step 4: Find the best nontabu move.
            best_move = None
            best_move_index = None
            best_objval_diff = float("inf")
            for i, move in enumerate(possible_moves):
                obj_val_diff = self.objective_func.update(
                    move.area, move.new_region, labels, attr)
                if obj_val_diff < best_objval_diff:
                    best_move_index, best_move = i, move
                    best_objval_diff = obj_val_diff
            # step 5: Make the move if possible. Update the tabu status.
            if self.allow_move_strategy(best_move.area, best_move.new_region, labels):
                self._make_move(best_move.area, best_move.new_region, labels, adj)
            # step 6: Look up the current zoning system in a list of all zoning
            # systems visited so far during the search. If not found then go
            # to step 10.
            # Sets can't be permuted so we convert our list to a set:
            label_tup = tuple(labels)
            if label_tup in self.visited:
                # step 7: If it is found and it has been visited more than K1
                # times already and this cyclical behavior has been found on
                # at least K2 other occasions (involving other zones) then go
                # to step 11.
                times_visited = self.visited.count(label_tup)
                cycle = list(reversed(self.visited))
                cycle = cycle[:cycle.index(label_tup) + 1]
                cycle = list(reversed(cycle))
                it_until_repetition = len(cycle)
                if times_visited > self.k1:
                    times_cycle_found = 0
                    if self.k2 > 0:
                        for i in range(len(self.visited) - len(cycle)):
                            if self.visited[i:i + len(cycle)] == cycle:
                                times_cycle_found += 1
                                if times_cycle_found >= self.k2:
                                    break
                    if times_cycle_found >= self.k2:
                        # step 11: Delete all stored zoning systems and make P
                        # random moves, P = 1 + self.avg_it_until_rep/2, and
                        # update tabu to preclude a return to the previous
                        # state.
                        # we save the labels such that we can access it if
                        # this step yields a poor solution.
                        last_step = (11, tuple(labels))
                        self.visited = []
                        p = math.floor(1 + self.avg_it_until_rep / 2)
                        possible_moves.pop(best_move_index)
                        for _ in range(p):
                            move = possible_moves.pop(random.randrange(len(possible_moves)))
                            if self.allow_move_strategy(move.area, move.new_region, labels):
                                self._make_move(move.area, move.new_region, labels, adj)
                        continue
                    # step 8: Update a moving average of the repetition
                    # interval self.avg_it_until_rep, and increase the
                    # prohibition period R to 1.1*R.
                    self.rep_counter += 1
                    avg_it = self.avg_it_until_rep
                    self.avg_it_until_rep = 1 / self.rep_counter * \
                        ((self.rep_counter-1)*avg_it + it_until_repetition)

                    self.tabu = deque(self.tabu, 1.1 * self.tabu.maxlen)
                    # step 9: If the number of iterations since R was last
                    # changed exceeds self.avg_it_until_rep, then decrease R to
                    # max(0.9*R, 1).
                    if it_since_tabu_len_changed > self.avg_it_until_rep:
                        new_tabu_len = max([0.9 * self.tabu.maxlen, 1])
                        new_tabu_len = math.floor(new_tabu_len)
                        self.tabu = deque(self.tabu, new_tabu_len)
                    it_since_tabu_len_changed = 0  # step 8

            # step 10: Save the zoning system and go to step 12.
            self.visited.append(tuple(labels))
            last_step = 10

        if last_step == 10:
            try:
                return np.array(self.visited[-2])
            except IndexError:
                return np.array(self.visited[-1])
        # if step 11 was the last one, the result is in last_step[1]
        return np.array(last_step[1])



import numpy as np
from geopandas import GeoDataFrame
from shapely.geometry import Polygon

from region.tests.util import region_list_from_array, convert_from_geodataframe
from region.util import dataframe_to_dict


# The following data describe the max-p-regions problem depicted in figure 2 on
# p. 402 in [DAR2012]_. They are commented out because the test cases take too
# long on Travis CI.
#attr = np.array([350.2, 400.5, 430.8,
#                 490.4, 410.9, 450.4,
#                 560.1, 500.7, 498.6])
#spatially_extensive_attr = np.array([30, 25, 31,
#                                     28, 32, 30,
#                                     35, 27, 33])
#threshold = 120
#optimal_clustering = region_list_from_array(np.array([0, 0, 0,
#                                                      1, 0, 0,
#                                                      1, 1, 1]))
#
#attr_str = "attr"
#spatially_extensive_attr_str = "spatially_extensive_attr"
#gdf = GeoDataFrame(
#        {attr_str: attr,
#         spatially_extensive_attr_str: spatially_extensive_attr},
#        geometry=[Polygon([(x, y),
#                           (x, y+1),
#                           (x+1, y+1),
#                           (x+1, y)]) for y in range(3) for x in range(3)]
#)


# New, more simple data (less areas than in figure 2 on p. 402 in [DAR2012]_).
# Using these data the tests run much faster than with the data from [DAR2012]_.
attr = np.array([0, 0,
                 1, 1])
# The following two lines mean that each region must consist of two regions.
# Since we have four areas, the optimal solution of the max-p problem
# will have 2 regions.
spatially_extensive_attr = np.ones(4)
threshold = 2
optimal_clustering = region_list_from_array(np.array([0, 0,
                                                      1, 1]))
attr_str = "attr"
spatially_extensive_attr_str = "spatially_extensive_attr"
gdf = GeoDataFrame(
        {attr_str: attr,
         spatially_extensive_attr_str: spatially_extensive_attr},
         geometry=[Polygon([(x, y),
                            (x, y+1),
                            (x+1, y+1),
                            (x+1, y)]) for y in range(2) for x in range(2)]
)

# for tests with scalar attr & spatially_extensive_attr per area
attr = attr.reshape(-1, 1)
spatially_extensive_attr = spatially_extensive_attr.reshape(-1, 1)

adj, graph, neighbors_dict, w = convert_from_geodataframe(gdf)
attr_dict = dataframe_to_dict(gdf, attr_str)
spatially_extensive_attr_dict = dataframe_to_dict(gdf,
                                                  spatially_extensive_attr_str)
# for tests where attr & spatially_extensive_attr are vectors in each area
double_attr = np.column_stack((attr, attr))
double_spatially_extensive_attr = np.column_stack((spatially_extensive_attr,
                                                   spatially_extensive_attr))
double_threshold = np.hstack((threshold, threshold))
double_attr_dict = dataframe_to_dict(gdf, [attr_str] * 2)
double_spatially_extensive_attr_dict = dataframe_to_dict(
        gdf, [spatially_extensive_attr_str] * 2)



import networkx as nx

from region.max_p_regions.heuristics import MaxPRegionsHeu
from region.tests.util import compare_region_lists, region_list_from_array

from region.max_p_regions.tests.data import (
    adj,
    neighbors_dict,
    gdf,
    graph,
    w,
    optimal_clustering,
)

# for tests with scalar attr and spatially_extensive_attr
from region.max_p_regions.tests.data import (
    attr_str,
    spatially_extensive_attr_str,
    attr,
    spatially_extensive_attr,
    threshold,
    attr_dict,
    spatially_extensive_attr_dict,
)

# for tests with non-scalar attr and spatially_extensive_attr
from region.max_p_regions.tests.data import (
    double_attr,
    double_spatially_extensive_attr,
    double_threshold,
    double_attr_dict,
    double_spatially_extensive_attr_dict,
)

attr = attr.reshape(-1)
spatially_extensive_attr = spatially_extensive_attr.reshape(-1)
# ### TESTS WITH SCALAR attr AND spatially_extensive_attr #####################
# test with csr_matrix
def test_scipy_sparse_matrix():
    cluster_object = MaxPRegionsHeu(random_state=0)
    cluster_object.fit_from_scipy_sparse_matrix(
        adj, attr, spatially_extensive_attr, threshold=threshold
    )
    obtained = region_list_from_array(cluster_object.labels_)
    compare_region_lists(obtained, optimal_clustering)


# tests with a GeoDataFrame
def test_geodataframe_basic():
    cluster_object = MaxPRegionsHeu(random_state=0)
    cluster_object.fit_from_geodataframe(
        gdf, attr_str, spatially_extensive_attr_str, threshold=threshold
    )
    obtained = region_list_from_array(cluster_object.labels_)
    compare_region_lists(obtained, optimal_clustering)


# tests with a dict as areas argument
def test_dict_basic():
    cluster_object = MaxPRegionsHeu(random_state=0)
    cluster_object.fit_from_dict(
        neighbors_dict, attr_dict, spatially_extensive_attr_dict, threshold=threshold
    )
    obtained = region_list_from_array(cluster_object.labels_)
    compare_region_lists(obtained, optimal_clustering)


# tests with Graph
# ... with dicts as attr and spatially_extensive_attr
def test_graph_dict_basic():
    cluster_object = MaxPRegionsHeu(random_state=0)
    cluster_object.fit_from_networkx(
        graph, attr_dict, spatially_extensive_attr_dict, threshold=threshold
    )
    result = region_list_from_array(cluster_object.labels_)
    compare_region_lists(result, optimal_clustering)


# ... with strings as attr and spatially_extensive_attr
def test_graph_str_basic():
    nx.set_node_attributes(graph, attr_dict, attr_str)
    nx.set_node_attributes(
        graph, spatially_extensive_attr_dict, spatially_extensive_attr_str
    )
    cluster_object = MaxPRegionsHeu(random_state=0)
    cluster_object.fit_from_networkx(
        graph, attr_str, spatially_extensive_attr_str, threshold=threshold
    )
    result = region_list_from_array(cluster_object.labels_)
    compare_region_lists(result, optimal_clustering)


# test with W
def test_w_basic():
    cluster_object = MaxPRegionsHeu(random_state=0)
    cluster_object.fit_from_w(w, attr, spatially_extensive_attr, threshold=threshold)
    result = region_list_from_array(cluster_object.labels_)
    compare_region_lists(result, optimal_clustering)


# ### TESTS WITH NON-SCALAR attr AND spatially_extensive_attr #################
# test with csr_matrix
def test_scipy_sparse_matrix_multi_attr():
    cluster_object = MaxPRegionsHeu(random_state=0)
    cluster_object.fit_from_scipy_sparse_matrix(
        adj, double_attr, double_spatially_extensive_attr, threshold=double_threshold
    )
    obtained = region_list_from_array(cluster_object.labels_)
    compare_region_lists(obtained, optimal_clustering)


# tests with a GeoDataFrame
def test_geodataframe_multi_attr():
    cluster_object = MaxPRegionsHeu(random_state=0)
    cluster_object.fit_from_geodataframe(
        gdf,
        [attr_str] * 2,
        [spatially_extensive_attr_str] * 2,
        threshold=double_threshold,
    )
    obtained = region_list_from_array(cluster_object.labels_)
    compare_region_lists(obtained, optimal_clustering)


# tests with a dict as areas argument
def test_dict_multi_attr():
    cluster_object = MaxPRegionsHeu(random_state=0)
    cluster_object.fit_from_dict(
        neighbors_dict,
        double_attr_dict,
        double_spatially_extensive_attr_dict,
        threshold=double_threshold,
    )
    obtained = region_list_from_array(cluster_object.labels_)
    compare_region_lists(obtained, optimal_clustering)


# tests with Graph
# ... with dicts as attr and spatially_extensive_attr
def test_graph_dict_multi_attr():
    cluster_object = MaxPRegionsHeu(random_state=0)
    cluster_object.fit_from_networkx(
        graph,
        double_attr_dict,
        double_spatially_extensive_attr_dict,
        threshold=double_threshold,
    )
    result = region_list_from_array(cluster_object.labels_)
    compare_region_lists(result, optimal_clustering)


# ... with strings as attr and spatially_extensive_attr
def test_graph_str_multi_attr():
    nx.set_node_attributes(graph, attr_dict, attr_str)
    nx.set_node_attributes(
        graph, spatially_extensive_attr_dict, spatially_extensive_attr_str
    )
    cluster_object = MaxPRegionsHeu(random_state=0)
    cluster_object.fit_from_networkx(
        graph,
        [attr_str] * 2,
        [spatially_extensive_attr_str] * 2,
        threshold=double_threshold,
    )
    result = region_list_from_array(cluster_object.labels_)
    compare_region_lists(result, optimal_clustering)


# test with W
def test_w_multi_attr():
    print(double_threshold)
    cluster_object = MaxPRegionsHeu(random_state=0)
    cluster_object.fit_from_w(
        w, double_attr, double_spatially_extensive_attr, threshold=double_threshold
    )
    result = region_list_from_array(cluster_object.labels_)
    compare_region_lists(result, optimal_clustering)



import random

import networkx as nx
from libpysal import weights

from region.objective_function import ObjectiveFunctionPairwise
from region.p_regions.azp import AZP
from region.p_regions.azp_util import AllowMoveAZPMaxPRegions
from region.util import array_from_df_col, array_from_dict_values, \
    array_from_graph_or_dict, array_from_region_list, copy_func, \
    find_sublist_containing, pop_randomly_from, raise_distance_metric_not_set,\
    random_element_from,scipy_sparse_matrix_from_w, \
    scipy_sparse_matrix_from_dict, w_from_gdf


class MaxPRegionsHeu:
    def __init__(self, local_search=None, random_state=None):
        """
        Class offering the implementation of the algorithm for solving the
        max-p-regions problem as described in [DAR2012]_.

        Parameters
        ----------
        local_search : None or :class:`AZP` or :class:`AZPSimulatedAnnealing`
            If None, then the AZP is used.
            Pass an instance of :class:`AZP` (or one of its subclasses) or
            :class:`AZPSimulatedAnnealing` to use a customized local search
            algorithm.
        random_state : None, int, str, bytes, or bytearray
            Random seed.
        """
        self.n_regions = None
        self.labels_ = None
        self.local_search = local_search
        self.random_state = random_state
        random.seed(random_state)
        self.metric = raise_distance_metric_not_set

    def fit_from_scipy_sparse_matrix(
            self,
            adj,
            attr,
            spatially_extensive_attr,
            threshold,
            max_it=10,
            objective_func=ObjectiveFunctionPairwise(),
            verbose=False):
        """
        Solve the max-p-regions problem in a heuristic way (see [DAR2012]_).

        The resulting region labels are assigned to the instance's
        :attr:`labels_` attribute.

        Parameters
        ----------
        adj : :class:`scipy.sparse.csr_matrix`
            Adjacency matrix representing the areas' contiguity relation.
        attr : :class:`numpy.ndarray`
            Array (number of areas x number of attributes) of areas' attributes
            relevant to clustering.
        spatially_extensive_attr : :class:`numpy.ndarray`
            Array (number of areas x number of attributes) of areas' attributes
            relevant to ensuring the threshold condition.
        threshold : numbers.Real or :class:`numpy.ndarray`
            The lower bound for a region's sum of spatially extensive
            attributes. The argument's type is numbers.Real if there is only
            one spatially extensive attribute per area, otherwise it is a
            one-dimensional array with as many entries as there are spatially
            extensive attributes per area.
        max_it : int, default: 10
            The maximum number of partitions produced in the algorithm's
            construction phase.
        objective_func : :class:`region.objective_function.ObjectiveFunction`, default: ObjectiveFunctionPairwise()
            The objective function to use.
        """

        w = weights.WSP(adj).to_W()
        areas_dict = w.neighbors
        self.metric = objective_func.metric

        best_partition = None
        best_obj_value = float("inf")
        feasible_partitions = []
        partitions_before_enclaves_assignment = []
        max_p = 0  # maximum number of regions

        # construction phase
        for _ in range(max_it):
            partition, enclaves = self.grow_regions(
                adj, attr, spatially_extensive_attr, threshold)
            n_regions = len(partition)
            if n_regions > max_p:
                partitions_before_enclaves_assignment = [(partition, enclaves)]
                max_p = n_regions
            elif n_regions == max_p:
                partitions_before_enclaves_assignment.append((partition,
                                                              enclaves))

        for partition, enclaves in partitions_before_enclaves_assignment:
            feasible_partitions.append(
                self.assign_enclaves(partition, enclaves, areas_dict, attr))

        for partition in feasible_partitions:

            # local search phase
            if self.local_search is None:
                self.local_search = AZP()
            self.local_search.allow_move_strategy = AllowMoveAZPMaxPRegions(
                spatially_extensive_attr, threshold,
                self.local_search.allow_move_strategy)
            for partition in feasible_partitions:
                self.local_search.fit_from_scipy_sparse_matrix(
                    adj,
                    attr,
                    max_p,
                    initial_labels=array_from_region_list(partition),
                    objective_func=objective_func)
                partition = self.local_search.labels_
                obj_value = objective_func(partition, attr)
                if obj_value < best_obj_value:
                    best_obj_value = obj_value
                    best_partition = partition
            self.labels_ = best_partition

    fit = copy_func(fit_from_scipy_sparse_matrix)
    fit.__doc__ = "Alias for :meth:`fit_from_scipy_sparse_matrix`.\n\n" \
                  + fit_from_scipy_sparse_matrix.__doc__

    def fit_from_dict(self,
                      neighbors_dict,
                      attr,
                      spatially_extensive_attr,
                      threshold,
                      max_it=10,
                      objective_func=ObjectiveFunctionPairwise()):
        """
        Solve the max-p-regions problem in a heuristic way (see [DAR2012]_).

        The resulting region labels are assigned to the instance's
        :attr:`labels_` attribute.

        Parameters
        ----------
        neighbors_dict : dict
            Each key represents an area and each value is an iterable of
            neighbors of this area.
        attr : dict
            A dict with the same keys as `neighbors_dict` and values
            representing the attributes for calculating homo-/heterogeneity. A
            value can be scalar (e.g. `float` or `int`) or a
            :class:`numpy.ndarray`.
        spatially_extensive_attr : dict
            A dict with the same keys as `neighbors_dict` and values
            representing the spatially extensive attribute (scalar or iterable
            of scalars). In the max-p-regions problem each region's sum of
            spatially extensive attributes must be greater than a specified
            threshold. In case of iterables of scalars as dict-values all
            elements of the iterable have to fulfill the condition.
        threshold : numbers.Real or :class:`numpy.ndarray`
            Refer to the corresponding argument in
            :meth:`fit_from_scipy_sparse_matrix`.
        max_it : int, default: 10
            Refer to the corresponding argument in
            :meth:`fit_from_scipy_sparse_matrix`.
        objective_func : :class:`region.ObjectiveFunction`, default: ObjectiveFunctionPairwise()
            Refer to the corresponding argument in
            :meth:`fit_from_scipy_sparse_matrix`.
        """
        if not isinstance(neighbors_dict, dict):
            raise ValueError("The neighbors_dict argument must be dict.")

        not_same_dict_keys_msg = "The {} argument has to be of type dict " \
                                 "with the same keys as neighbors_dict."

        if not isinstance(attr, dict) or attr.keys() != neighbors_dict.keys():
            raise ValueError(not_same_dict_keys_msg.format("attr"))

        if not isinstance(spatially_extensive_attr, dict) or \
                spatially_extensive_attr.keys() != neighbors_dict.keys():
            raise ValueError(
                not_same_dict_keys_msg.format(spatially_extensive_attr))
        adj = scipy_sparse_matrix_from_dict(neighbors_dict)
        attr_arr = array_from_dict_values(attr)
        spat_ext_attr_arr = array_from_dict_values(spatially_extensive_attr)
        self.fit_from_scipy_sparse_matrix(
            adj,
            attr_arr,
            spat_ext_attr_arr,
            threshold=threshold,
            max_it=max_it,
            objective_func=objective_func)

    def fit_from_geodataframe(self,
                              gdf,
                              attr,
                              spatially_extensive_attr,
                              threshold,
                              max_it=10,
                              objective_func=ObjectiveFunctionPairwise(),
                              contiguity="rook"):
        """
        Alternative API for :meth:`fit_from_scipy_sparse_matrix`.

        Parameters
        ----------
        gdf : :class:`geopandas.GeoDataFrame`

        attr : str or list
            The clustering criteria (columns of the GeoDataFrame `gdf`) are
            specified as string (for one column) or list of strings (for
            multiple columns).
        spatially_extensive_attr : str or list
            The name (`str`) or names (`list` of strings) of column(s) in `gdf`
            containing the spatially extensive attributes.
        threshold : numbers.Real or :class:`numpy.ndarray`
            Refer to the corresponding argument in
            :meth:`fit_from_scipy_sparse_matrix`.
        max_it : int, default: 10
            Refer to the corresponding argument in
            :meth:`fit_from_scipy_sparse_matrix`.
        objective_func : :class:`region.ObjectiveFunction`, default: ObjectiveFunctionPairwise()
            Refer to the corresponding argument in
            :meth:`fit_from_scipy_sparse_matrix`.
        contiguity : {"rook", "queen"}, default: "rook"
            Defines the contiguity relationship between areas. Possible
            contiguity definitions are:

            * "rook" - Rook contiguity.
            * "queen" - Queen contiguity.
        """
        w = w_from_gdf(gdf, contiguity)
        attr = array_from_df_col(gdf, attr)
        spat_ext_attr = array_from_df_col(gdf, spatially_extensive_attr)

        self.fit_from_w(
            w,
            attr,
            spat_ext_attr,
            threshold=threshold,
            max_it=max_it,
            objective_func=objective_func)

    def fit_from_networkx(self,
                          graph,
                          attr,
                          spatially_extensive_attr,
                          threshold,
                          max_it=10,
                          objective_func=ObjectiveFunctionPairwise()):
        """
        Alternative API for :meth:`fit_from_scipy_sparse_matrix`.

        Parameters
        ----------
        graph : `networkx.Graph`

        attr : str, list or dict
            If the clustering criteria are present in the networkx.Graph
            `graph` as node attributes, then they can be specified as a string
            (for one criterion) or as a list of strings (for multiple
            criteria).
            Alternatively, a dict can be used with each key being a node of the
            networkx.Graph `graph` and each value being the corresponding
            clustering criterion (a scalar (e.g. `float` or `int`) or a
            :class:`numpy.ndarray`).
            If there are no clustering criteria present in the networkx.Graph
            `graph` as node attributes, then a dictionary must be used for this
            argument. Refer to the corresponding argument in
            :meth:`fit_from_dict` for more details about the expected dict.
        spatially_extensive_attr : str, list or dict
            If the spatially extensive attribute is present in the
            networkx.Graph `graph` as node attributes, then they can be
            specified as a string (for one attribute) or as a list of
            strings (for multiple attributes).
            Alternatively, a dict can be used with each key being a node of the
            networkx.Graph `graph` and each value being the corresponding
            spatially extensive attribute (a scalar (e.g. `float` or `int`) or
            a :class:`numpy.ndarray`).
            If there are no spatially extensive attributes present in the
            networkx.Graph `graph` as node attributes, then a dictionary must
            be used for this argument. Refer to the corresponding argument in
            :meth:`fit_from_dict` for more details about the expected dict.
        threshold : numbers.Real or :class:`numpy.ndarray`
            Refer to the corresponding argument in
            :meth:`fit_from_scipy_sparse_matrix`.
        max_it : int, default: 10
            Refer to the corresponding argument in
            :meth:`fit_from_scipy_sparse_matrix`.
        objective_func : :class:`region.ObjectiveFunction`, default: ObjectiveFunctionPairwise()
            Refer to the corresponding argument in
            :meth:`fit_from_scipy_sparse_matrix`.
        """
        adj = nx.to_scipy_sparse_matrix(graph)
        attr = array_from_graph_or_dict(graph, attr)
        sp_ext_attr = array_from_graph_or_dict(graph, spatially_extensive_attr)
        self.fit_from_scipy_sparse_matrix(
            adj,
            attr,
            sp_ext_attr,
            threshold=threshold,
            max_it=max_it,
            objective_func=objective_func)

    def fit_from_w(self,
                   w,
                   attr,
                   spatially_extensive_attr,
                   threshold,
                   max_it=10,
                   objective_func=ObjectiveFunctionPairwise()):
        """
        Alternative API for :meth:`fit_from_scipy_sparse_matrix`.

        Parameters
        ----------
        w : :class:`libpysal.weights.weights.W`
            W object representing the contiguity relation.
        attr : :class:`numpy.ndarray`
            Each element specifies an area's attribute which is used for
            calculating the objective function.
        spatially_extensive_attr : :class:`numpy.ndarray`
            Each element specifies an area's spatially extensive attribute
            which is used to ensure that the sum of spatially extensive
            attributes in each region adds up to a threshold defined by the
            `threshold` argument.
        threshold : numbers.Real or :class:`numpy.ndarray`
            Refer to the corresponding argument in
            :meth:`fit_from_scipy_sparse_matrix`.
        max_it : int, default: 10
            Refer to the corresponding argument in
            :meth:`fit_from_scipy_sparse_matrix`.
        objective_func : :class:`region.ObjectiveFunction`, default: ObjectiveFunctionPairwise()
            Refer to the corresponding argument in
            :meth:`fit_from_scipy_sparse_matrix`.
        """
        adj = scipy_sparse_matrix_from_w(w)
        self.fit_from_scipy_sparse_matrix(
            adj,
            attr,
            spatially_extensive_attr,
            threshold,
            max_it=max_it,
            objective_func=objective_func)

    def grow_regions(self, adj, attr, spatially_extensive_attr, threshold):
        """
        Parameters
        ----------
        adj : :class:`scipy.sparse.csr_matrix`
            See the corresponding argument in
            :meth:`fit_from_scipy_sparse_matrix`.
        attr : :class:`numpy.ndarray`
            See the corresponding argument in
            :meth:`fit_from_scipy_sparse_matrix`.
        spatially_extensive_attr : :class:`numpy.ndarray`
            See the corresponding argument in
            :meth:`fit_from_scipy_sparse_matrix`.
        threshold : numbers.Real or :class:`numpy.ndarray`
            See the corresponding argument in
            :meth:`fit_from_scipy_sparse_matrix`.

        Returns
        -------
        result : `tuple`
            `result[0]` is a `list`. Each list element is a `set` of a region's
            areas. Note that not every area is assigned to a region after this
            function has terminated, so they won't be in any of the `set`s in
            `result[0]`.
            `result[1]` is a `list` of areas not assigned to any region.
        """
        partition = []
        enclave_areas = []
        unassigned_areas = list(range(adj.shape[0]))
        assigned_areas = []

        while unassigned_areas:
            area = pop_randomly_from(unassigned_areas)
            assigned_areas.append(area)
            if (spatially_extensive_attr[area] >= threshold).all():
                partition.append({area})
            else:
                region = {area}
                unassigned_neighs = set(
                    adj[area].nonzero()[1]).difference(assigned_areas)
                feasible = True
                spat_ext_attr = spatially_extensive_attr[area].copy()
                while (spat_ext_attr < threshold).any():
                    if unassigned_neighs:
                        neigh = self.find_best_area(region, unassigned_neighs,
                                                    attr)
                        region.add(neigh)
                        unassigned_neighs.remove(neigh)
                        unassigned_neighs.update(set(adj[neigh].nonzero()[1]))
                        unassigned_neighs.difference_update(assigned_areas)
                        spat_ext_attr += spatially_extensive_attr[neigh]
                        unassigned_areas.remove(neigh)
                        assigned_areas.append(neigh)
                    else:
                        enclave_areas.extend(region)
                        feasible = False
                        # the following line (present in the algorithm in
                        # [DAR2012]) is commented out because it leads to an
                        # infinite loop:
                        # unassigned_areas.extend(region)
                        for area in region:
                            assigned_areas.remove(area)
                        break
                if feasible:
                    partition.append(region)
        return partition, enclave_areas

    def find_best_area(self, region, candidates, attr):
        """
        Parameters
        ----------
        region : iterable
            Each element represents an area.
        candidates : iterable
            Each element represents an area bordering on region.
        attr : :class:`numpy.ndarray`
            See the corresponding argument in
            :meth:`fit_from_scipy_sparse_matrix`.

        Returns
        -------
        best_area :
            An element of `candidates` with minimal dissimilarity when being
            moved to the region `region`.
        """
        candidates = {
            area: sum(
                self.metric(attr[area].reshape(1, -1), attr[area2].reshape(
                    1, -1)) for area2 in region)
            for area in candidates
        }
        best_candidates = [
            area for area in candidates
            if candidates[area] == min(candidates.values())
        ]
        return random_element_from(best_candidates)

    def assign_enclaves(self, partition, enclave_areas, neigh_dict, attr):
        """
        Start with a partial partition (not all areas are assigned to a region)
        and a list of enclave areas (i.e. areas not present in the partial
        partition). Then assign all enclave areas to regions in the partial
        partition and return the resulting partition.

        Parameters
        ----------
        partition : `list`
            Each element (of type `set`) represents a region.
        enclave_areas : `list`
            Each element represents an area.
        neigh_dict : `dict`
            Each key represents an area. Each value is an iterable of the
            corresponding neighbors.
        attr : :class:`numpy.ndarray`
            See the corresponding argument in
            :meth:`fit_from_scipy_sparse_matrix`.

        Returns
        -------
        partition : `list`
            Each element (of type `set`) represents a region.
        """
        while enclave_areas:
            neighbors_of_assigned = [
                area for area in enclave_areas if any(
                    neigh not in enclave_areas for neigh in neigh_dict[area])
            ]
            area = pop_randomly_from(neighbors_of_assigned)
            neigh_regions_idx = []
            for neigh in neigh_dict[area]:
                try:
                    neigh_regions_idx.append(
                        find_sublist_containing(neigh, partition, index=True))
                except LookupError:
                    pass
            region_idx = self.find_best_region_idx(area, partition,
                                                   neigh_regions_idx, attr)
            partition[region_idx].add(area)
            enclave_areas.remove(area)
        return partition

    def find_best_region_idx(self, area, partition, candidate_regions_idx,
                             attr):
        """

        Parameters
        ----------
        area :
            The area to be moved to one of the regions specified by
            `candidate_regions_idx`.
        partition : `list`
            Each element (of type `set`) represents a region.
        candidate_regions_idx : iterable
            Each element is the index of a region in the `partition` list.
        attr : :class:`numpy.ndarray`
            See the corresponding argument in
            :meth:`fit_from_scipy_sparse_matrix`.

        Returns
        -------
        best_idx : int
            The index of a region (w.r.t. `partition`) which has the smallest
            sum of dissimilarities after area `area` is moved to the region.
        """
        dissim_per_idx = {
            region_idx: sum(
                self.metric(attr[area].reshape(1, -1), attr[area2].reshape(
                    1, -1)) for area2 in partition[region_idx])
            for region_idx in candidate_regions_idx
        }
        minimum_dissim = min(dissim_per_idx.values())
        best_idxs = [
            idx for idx in dissim_per_idx
            if dissim_per_idx[idx] == minimum_dissim
        ]
        return random_element_from(best_idxs)



import abc
import math
import numbers
import random

import numpy as np


class AllowMoveStrategy(abc.ABC):
    def start_new_component(self, initial_labels, attr, objective_func,
                            comp_idx):
        """
        This method should be called whenever a new connected component is
        clustered.

        Parameters
        ----------
        initial_labels : :class:`numpy.ndarray`
            The region labels of the areas in the currently considered
            connected component.
            Shape: number of areas in the currently considered component.
        attr : :class:`numpy.ndarray`
            The areas' attributes.
            Shape: number of areas in the currently considered component.
        objective_func : :class:`region.objective_function.ObjectiveFunction`
            The objective function to use.
        comp_idx : :class:`numpy.ndarray`
            The indices of those areas belonging to the component clustered
            next.
        """
        self.attr = attr
        self.objective_func = objective_func
        self.objective_val = self.objective_func(initial_labels, self.attr)

    @abc.abstractmethod
    def __call__(self, moving_area, new_region, labels):
        """
        Assess whether a potential move is allowed or not. The criteria for
        allowing a move are defined in the implementations of this abstract
        method (in subclasses).

        Parameters
        ----------
        moving_area : int
            Area involved in a potential move.
        new_region : int
            Region to which the area `moving_area` is moved if the move is
            allowed.
        labels : :class:`numpy.ndarray`
            Region labels of areas in the currently considered connected
            component.

        Returns
        -------
        is_allowed : `bool`
            `True` if area `moving_area` is allowed to move to `new_region`.
            `False` otherwise.
        """


class AllowMoveAZP(AllowMoveStrategy):
    def __call__(self, moving_area, new_region, labels):
        diff = self.objective_func.update(moving_area, new_region, labels,
                                          self.attr)
        if diff <= 0:
            self.objective_val += diff
            return True
        else:

            return False


class AllowMoveAZPSimulatedAnnealing(AllowMoveStrategy):
    def __init__(self, init_temperature, sa_moves_term=float("inf")):
        self.observers_min_sa_moves = []
        self.observers_move_made = []
        self.t = init_temperature
        if not isinstance(sa_moves_term, numbers.Integral) or \
                sa_moves_term < 1:
            raise ValueError("The sa_moves_term argument must be a positive "
                             "integer.")
        self.sa_moves_term = sa_moves_term
        self.sa = 0  # number of SA-moves
        super().__init__()

    def __call__(self, moving_area, new_region, labels):
        diff = self.objective_func.update(moving_area, new_region, labels,
                                          self.attr)
        if diff <= 0:
            self.objective_val += diff
            self.notify_move_made()
            return True
        else:

            prob = math.exp(-diff / self.t)
            move_allowed = random.random() < prob
            if move_allowed:
                self.notify_move_made()
                self.sa += 1
                if self.sa >= self.sa_moves_term:
                    self.notify_min_sa_moves()
                self.objective_val += diff
                return True
            return False

    def register_sa_moves_term(self, observer_func):
        """
        Parameters
        ----------
        observer_func : callable
            A function to call when the certain number of SA-moves has been
            reached. This number is called Q in [OR1995]_ (page 431).
        """
        if callable(observer_func):
            self.observers_min_sa_moves.append(observer_func)
        else:
            raise ValueError("The observer_func must be callable.")

    def register_move_made(self, observer_func):
        """
        Parameters
        ----------
        observer_func : callable
            A function to call when a move is allowed.
        """
        if callable(observer_func):
            self.observers_move_made.append(observer_func)
        else:
            raise ValueError("The observer_func must be callable.")

    def notify_min_sa_moves(self):
        for observer_func in self.observers_min_sa_moves:
            observer_func()

    def notify_move_made(self):
        for observer_func in self.observers_move_made:
            observer_func()

    def update_temperature(self, temp):
        self.t = temp

    def reset(self):
        self.sa = 0  # number of SA-moves


class AllowMoveAZPMaxPRegions(AllowMoveStrategy):
    """
    Ensures that the spatially extensive attribute adds up to a
    given threshold in each region. Only moves preserving this condition in
    both the donor as well as the recipient region are allowed. The check for
    the recipient region is necessary in case there is an area with a negative
    spatially extensive attribute.
    """

    def __init__(self, spatially_extensive_attr, threshold,
                 decorated_strategy):
        """

        Parameters
        ----------
        spatially_extensive_attr : :class:`numpy.ndarray`, default: None
            See corresponding argument in
            :meth:`region.max_p_regions.heuristics.MaxPRegionsHeu.fit_from_scipy_sparse_matrix`.
        threshold : numbers.Real or :class:`numpy.ndarray`
            See corresponding argument in
            :meth:`region.max_p_regions.heuristics.MaxPRegionsHeu.fit_from_scipy_sparse_matrix`
        decorated_strategy : :class:`AllowMoveStrategy`
            The :class:`AllowMoveStrategy` related to the algorithms local
            search.
        """
        self._decorated_strategy = decorated_strategy
        self.spatially_extensive_attr_all = spatially_extensive_attr
        self.spatially_extensive_attr = None
        self.threshold = threshold

    def start_new_component(self, initial_labels, attr, objective_func,
                            comp_idx):
        self.spatially_extensive_attr = self.spatially_extensive_attr_all[
            comp_idx]
        super().start_new_component(initial_labels, attr, objective_func,
                                    comp_idx)
        self._decorated_strategy.start_new_component(initial_labels, attr,
                                                     objective_func, comp_idx)

    def __call__(self, moving_area, new_region, labels):
        sp_ext = self.spatially_extensive_attr

        if (sp_ext[moving_area]).any() > 0:
            donor_region = labels[moving_area]
            donor_idx = np.where(labels == donor_region)[0]
            donor_sum = sum(sp_ext[donor_idx]) - sp_ext[moving_area]
            threshold_reached_donor = (donor_sum >= self.threshold).all()
            if not threshold_reached_donor:
                return False

        elif (sp_ext[moving_area]).any() < 0:
            recipient_idx = np.where(labels == new_region)[0]
            recipient_sum = sum(sp_ext[recipient_idx]) + sp_ext[moving_area]
            threshold_reached_recipient = (recipient_sum >=
                                           self.threshold).all()
            if not threshold_reached_recipient:
                return False

        return self._decorated_strategy(moving_area, new_region, labels)

    def __getattr__(self, name):
        """
        Forward calls to unimplemented methods to _decorated_strategy
        (necessary e.g. for :class:`AllowMoveAZPSimulatedAnnealing`).
        """
        return getattr(self._decorated_strategy, name)



from scipy.sparse import csr_matrix
import numpy as np
import pytest

from region.p_regions.azp import AZP
from region import util


def all_elements_equal(iterable):
    return all(iterable[0] == element for element in iterable)


def not_all_elements_equal(iterable):
    return not all_elements_equal(iterable)


def test_assert_feasible__pass_disconnected():
    adj = csr_matrix(np.array([[0, 0],
                               [0, 0]]))  # nodes 0 and 1 are not connected ...
    labels = np.array([0, 1])  # ... and assigned to different regions
    try:
        util.assert_feasible(labels, adj)
        util.assert_feasible(labels, adj, n_regions=2)
    except ValueError:
        pytest.fail()


def test_assert_feasible__pass_connected():
    adj = csr_matrix(np.array([[0, 1],
                               [1, 0]]))  # nodes 0 and 1 are connected ...
    labels = np.array([0, 0])  # ...and (case 1) assigned to the same region
    try:
        util.assert_feasible(labels, adj)
        util.assert_feasible(labels, adj, n_regions=1)
    except ValueError:
        pytest.fail()

    labels = np.array([0, 1])  # ...and (case 2) assigned to different regions
    try:
        util.assert_feasible(labels, adj)
        util.assert_feasible(labels, adj, n_regions=2)
    except ValueError:
        pytest.fail()


def test_assert_feasible__contiguity():
    with pytest.raises(ValueError) as exc_info:
        # nodes 0 and 1 are not connected ...
        adj = csr_matrix(np.array([[0, 0],
                                   [0, 0]]))
        # ... but assigned to the same region --> not feasible
        labels = np.array([0, 0])
        util.assert_feasible(labels, adj)
    assert "not spatially contiguous" in str(exc_info.value)


def test_assert_feasible__number_of_regions():
    with pytest.raises(ValueError) as exc_info:
        adj = csr_matrix(np.array([[0, 1],
                                   [1, 0]]))  # nodes 0 and 1 are connected ...
        labels = np.array([0, 0])  # ... and assigned to the same region
        # but this is not feasible under the condition n_regions = 2
        n_regions = 2
        util.assert_feasible(labels, adj, n_regions=n_regions)
    assert "The number of regions is" in str(exc_info.value)


def test_random_element_from():
    lst = list(range(1000))
    n_pops = 5
    popped = []
    for _ in range(n_pops):
        lst_copy = list(lst)
        popped.append(util.random_element_from(lst_copy))
        assert len(lst_copy) == len(lst)
    assert not_all_elements_equal(popped)


def test_pop_randomly_from():
    lst = list(range(1000))
    n_pops = 5
    popped = []
    for _ in range(n_pops):
        lst_copy = list(lst)
        popped.append(util.pop_randomly_from(lst_copy))
        assert len(lst_copy) == len(lst) - 1
    assert not_all_elements_equal(popped)


def test_AZP_azp_connected_component__one_area():
    adj = csr_matrix(np.array([[0]]))  # adjacency matrix for a single node
    azp = AZP()
    obtained = azp._azp_connected_component(adj,
                                            initial_clustering=np.array([0]),
                                            attr=np.array([123]))
    desired = np.array([0])
    assert obtained == desired



import collections
import functools
import itertools
import random
import types

import scipy.sparse.csgraph as csg
from sklearn.metrics.pairwise import distance_metrics
from scipy.sparse.dok import dok_matrix
import numpy as np
import networkx as nx
from libpysal import weights
import pulp

from region.csgraph_utils import sub_adj_matrix, is_connected

Move = collections.namedtuple("move", "area old_region new_region")
"A named tuple representing a move from `old_region` to `new_region`."  # sphinx


def array_from_dict_values(dct, sorted_keys=None, flat_output=False, dtype=np.float):
    """
    Return values of the dictionary passed as `dct` argument as an numpy array.
    The values in the returned array are sorted by the keys of `dct`.

    Parameters
    ----------
    dct : dict

    sorted_keys : iterable, optional
        If passed, then the elements of the returned array will be sorted by
        this argument. Thus, this argument can be passed to suppress the
        sorting, or for getting a subset of the dictionary's values or to get
        repeated values.
    flat_output : bool, default: False
        If True, the returned array will be one-dimensional.
        If False, the returned array will be two-dimensional with one row per
        key in `dct`.
    dtype : default: np.float64
        The `dtype` of the returned array.

    Returns
    -------
    array : :class:`numpy.ndarray`

    Examples
    --------
    >>> dict_flat = {0: 0, 1: 10}
    >>> dict_it = {0: [0], 1: [10]}
    >>> desired_flat = np.array([0, 10])
    >>> desired_2d = np.array([[0],
    ...                        [10]])
    >>> flat_flat = array_from_dict_values(dict_flat, flat_output=True)
    >>> (flat_flat == desired_flat).all()
    True
    >>> flat_2d = array_from_dict_values(dict_flat)
    >>> (flat_2d == desired_2d).all()
    True
    >>> it_flat = array_from_dict_values(dict_it, flat_output=True)
    >>> (it_flat == desired_flat).all()
    True
    >>> it_2d = array_from_dict_values(dict_it)
    >>> (it_2d == desired_2d).all()
    True
    """
    if sorted_keys is None:
        sorted_keys = sorted(dct)
    iterable_values = isinstance(dct[sorted_keys[0]], collections.Iterable)
    if iterable_values:
        it = itertools.chain.from_iterable(dct[key] for key in sorted_keys)
    else:
        it = (dct[key] for key in sorted_keys)

    flat_arr = np.fromiter(it, dtype=dtype)
    if flat_output:
        return flat_arr
    return flat_arr.reshape((len(dct), -1))


def scipy_sparse_matrix_from_dict(neighbors):
    """
    Parameters
    ----------
    neighbors : dict
        Each key represents an area. The corresponding value contains the
        area's neighbors.

    Returns
    -------
    adj : :class:`scipy.sparse.csr_matrix`
        Adjacency matrix representing the areas' contiguity relation.

    Examples
    --------
    >>> neighbors = {0: {1, 3}, 1: {0, 2, 4}, 2: {1, 5},
    ...              3: {0, 4}, 4: {1, 3, 5}, 5: {2, 4}}
    >>> obtained = scipy_sparse_matrix_from_dict(neighbors)
    >>> desired = np.array([[0, 1, 0, 1, 0, 0],
    ...                     [1, 0, 1, 0, 1, 0],
    ...                     [0, 1, 0, 0, 0, 1],
    ...                     [1, 0, 0, 0, 1, 0],
    ...                     [0, 1, 0, 1, 0, 1],
    ...                     [0, 0, 1, 0, 1, 0]])
    >>> (obtained.todense() == desired).all()
    True
    >>> neighbors = {"left": {"middle"},
    ...              "middle": {"left", "right"},
    ...              "right": {"middle"}}
    >>> obtained = scipy_sparse_matrix_from_dict(neighbors)
    >>> desired = np.array([[0, 1, 0],
    ...                     [1, 0, 1],
    ...                     [0, 1, 0]])
    >>> (obtained.todense() == desired).all()
    True
    """
    n_areas = len(neighbors)
    name_to_int = {area_name: i for i, area_name in enumerate(sorted(neighbors))}
    adj = dok_matrix((n_areas, n_areas))
    for i in neighbors:
        for j in neighbors[i]:
            adj[name_to_int[i], name_to_int[j]] = 1
    return adj.tocsr()


def scipy_sparse_matrix_from_w(w):
    """

    Parameters
    ----------
    w : :class:`libpysal.weights.weights.W`
        A W object representing the areas' contiguity relation.

    Returns
    -------
    adj : :class:`scipy.sparse.csr_matrix`
        Adjacency matrix representing the areas' contiguity relation.

    Examples
    --------
    >>> from libpysal import weights
    >>> neighbor_dict = {0: {1}, 1: {0, 2}, 2: {1}}
    >>> w = weights.W(neighbor_dict)
    >>> obtained = scipy_sparse_matrix_from_w(w)
    >>> desired = np.array([[0., 1., 0.],
    ...                     [1., 0., 1.],
    ...                     [0., 1., 0.]])
    >>> obtained.todense().all() == desired.all()
    True
    """
    return w.sparse


def dict_from_graph_attr(graph, attr, array_values=False):
    """
    Parameters
    ----------
    graph : networkx.Graph

    attr : str, iterable, or dict
        If str, then it specifies the an attribute of the graph's nodes.
        If iterable of strings, then multiple attributes of the graph's nodes
        are specified.
        If dict, then each key is a node and each value the corresponding
        attribute value. (This format is also this function's return format.)
    array_values : bool, default: False
        If True, then each value is transformed into a :class:`numpy.ndarray`.

    Returns
    -------
    result_dict : dict
        Each key is a node in the graph.
        If `array_values` is False, then each value is a list of attribute
        values corresponding to the key node.
        If `array_values` is True, then each value this list of attribute
        values is turned into a :class:`numpy.ndarray`. That requires the
        values to be shape-compatible for stacking.

    Examples
    --------
    >>> import networkx as nx
    >>> edges = [(0, 1), (1, 2),          # 0 | 1 | 2
    ...          (0, 3), (1, 4), (2, 5),  # ---------
    ...          (3, 4), (4,5)]           # 3 | 4 | 5
    >>> graph = nx.Graph(edges)
    >>> data_dict = {node: 10*node for node in graph}
    >>> nx.set_node_attributes(graph, data_dict, "test_data")
    >>> desired = {key: [value] for key, value in data_dict.items()}
    >>> dict_from_graph_attr(graph, "test_data") == desired
    True
    >>> dict_from_graph_attr(graph, ["test_data"]) == desired
    True
    """
    if isinstance(attr, dict):
        return attr
    if isinstance(attr, str):
        attr = [attr]
    data_dict = {node: [] for node in graph.nodes()}
    for a in attr:
        for node, value in nx.get_node_attributes(graph, a).items():
            data_dict[node].append(value)
    if array_values:
        for node in data_dict:
            data_dict[node] = np.array(data_dict[node])
    return data_dict


def array_from_graph(graph, attr):
    """

    Parameters
    ----------
    graph : networkx.Graph

    attr : str or iterable
        If str, then it specifies the an attribute of the graph's nodes.
        If iterable of strings, then multiple attributes of the graph's nodes
        are specified.

    Returns
    -------
    array : :class:`numpy.ndarray`
        Array with one row for each node in `graph`.

    Examples
    --------
    >>> import networkx as nx
    >>> edges = [(0, 1), (1, 2),          # 0 | 1 | 2
    ...          (0, 3), (1, 4), (2, 5),  # ---------
    ...          (3, 4), (4,5)]           # 3 | 4 | 5
    >>> graph = nx.Graph(edges)
    >>> data_dict = {node: 10*node for node in graph}
    >>> nx.set_node_attributes(graph, data_dict, "test_data")
    >>> desired = np.array([[0],
    ...                     [10],
    ...                     [20],
    ...                     [30],
    ...                     [40],
    ...                     [50]])
    >>> (array_from_graph(graph, "test_data") == desired).all()
    True
    >>> (array_from_graph(graph, ["test_data"]) == desired).all()
    True
    >>> (array_from_graph(graph, ["test_data", "test_data"]) ==
    ...  np.hstack((desired, desired))).all()
    True
    """
    dct = dict_from_graph_attr(graph, attr)
    return array_from_dict_values(dct)


def array_from_graph_or_dict(graph, attr):
    if isinstance(attr, (str, collections.Iterable)):
        return array_from_graph(graph, attr)
    elif isinstance(attr, collections.Mapping):
        return array_from_dict_values(attr)
    else:
        raise ValueError(
            "The `attr` argument must be a string, a list of "
            "strings or a dictionary."
        )


def array_from_region_list(region_list):
    """
    Parameters
    ----------
    region_list : `list`
        Each list element is an iterable of a region's areas.

    Returns
    -------
    labels : :class:`numpy.ndarray`
        Each element specifies the region of the corresponding area.

    Examples
    --------
    >>> import numpy as np
    >>> obtained = array_from_region_list([{0, 1, 2, 5}, {3, 4}])
    >>> desired = np.array([ 0, 0, 0, 1, 1, 0])
    >>> (obtained == desired).all()
    True
    """
    n_areas = sum(len(region) for region in region_list)
    labels = np.zeros((n_areas))
    for region_idx, region in enumerate(region_list):
        for area in region:
            labels[area] = region_idx
    return labels


def array_from_df_col(df, attr):
    """
    Extract one or more columns from a DataFrame as numpy array.

    Parameters
    ----------
    df : Union[DataFrame, GeoDataFrame]

    attr : Union[str, Sequence[str]]
        The columns' names to extract.

    Returns
    -------
    col : :class:`numpy.ndarray`
        The specified column(s) of the array.

    Examples
    --------
    >>> import pandas as pd
    >>> df = pd.DataFrame({"col1": [1, 2, 3],
    ...                    "col2": [7, 8, 9]})
    >>> (array_from_df_col(df, "col1") == np.array([[1],
    ...                                         [2],
    ...                                         [3]])).all()
    True
    >>> (array_from_df_col(df, ["col1"]) == np.array([[1],
    ...                                           [2],
    ...                                           [3]])).all()
    True
    >>> (array_from_df_col(df, ["col1", "col2"]) == np.array([[1, 7],
    ...                                                   [2, 8],
    ...                                                   [3, 9]])).all()
    True
    """
    value_error = ValueError(
        "The attr argument has to be of one of the "
        "following types: str or a sequence of strings."
    )
    if isinstance(attr, str):
        attr = [attr]
    elif isinstance(attr, collections.Sequence):
        if not all(isinstance(el, str) for el in attr):
            raise value_error
    else:
        raise value_error
    return np.array(df[attr])


def w_from_gdf(gdf, contiguity):
    """
    Get a `W` object from a GeoDataFrame.

    Parameters
    ----------
    gdf : GeoDataFrame

    contiguity : {"rook", "queen"}

    Returns
    -------
    cweights : `W`
        The contiguity information contained in the `gdf` argument in the form
        of a W object.
    """
    if not isinstance(contiguity, str) or contiguity.lower() not in ["rook", "queen"]:
        raise ValueError(
            "The contiguity argument must be either None "
            "or one of the following strings: "
            '"rook" or"queen".'
        )
    if contiguity.lower() == "rook":
        cweights = weights.Rook.from_dataframe(gdf)
    else:  # contiguity.lower() == "queen"
        cweights = weights.Queen.from_dataframe(gdf)
    return cweights


def dataframe_to_dict(df, cols):
    """
    Parameters
    ----------
    df : Union[:class:`pandas.DataFrame`, :class:`geopandas.GeoDataFrame`]

    cols : Union[`str`,  `list`]
        If `str`, then it is the name of a column of `df`.
        If `list`, then it is a list of strings. Each string is the name of a
        column of `df`.

    Returns
    -------
    result : dict
        The keys are the elements of the DataFrame's index.
        Each value is a :class:`numpy.ndarray` holding the corresponding values
        in the columns specified by `cols`.

    Examples
    --------
    >>> import pandas as pd
    >>> df = pd.DataFrame({"data": [100, 120, 115]})
    >>> result = dataframe_to_dict(df, "data")
    >>> result == {0: 100, 1: 120, 2: 115}
    True
    >>> import numpy as np
    >>> df = pd.DataFrame({"data": [100, 120],
    ...                    "other": [1, 2]})
    >>> actual = dataframe_to_dict(df, ["data", "other"])
    >>> desired = {0: np.array([100, 1]), 1: np.array([120, 2])}
    >>> all(np.array_equal(actual[i], desired[i]) for i in desired)
    True
    """
    return dict(zip(df.index, np.array(df[cols])))


def find_sublist_containing(el, lst, index=False):
    """

    Parameters
    ----------
    el :
        The element to search for in the sublists of `lst`.
    lst : collections.Sequence
        A sequence of sequences or sets.
    index : bool, default: False
        If False (default), the subsequence or subset containing `el` is
        returned.
        If True, the index of the subsequence or subset in `lst` is returned.

    Returns
    -------
    result : collections.Sequence, collections.Set or int
        See the `index` argument for more information.

    Raises
    ------
    exc : LookupError
        If `el` is not in any of the elements of `lst`.

    Examples
    --------
    >>> lst = [{0, 1}, {2}]
    >>> find_sublist_containing(0, lst, index=False) == {0, 1}
    True
    >>> find_sublist_containing(0, lst, index=True) == 0
    True
    >>> find_sublist_containing(2, lst, index=False) == {2}
    True
    >>> find_sublist_containing(2, lst, index=True) == 1
    True
    """
    for idx, sublst in enumerate(lst):
        if el in sublst:
            return idx if index else sublst
    raise LookupError("{} not found in any of the sublists of {}".format(el, lst))


def get_metric_function(metric=None):
    """
    Parameters
    ----------
    metric : str or function or None, default: None
        Using None is equivalent to using "euclidean".

        If str, then this string specifies the distance metric (from
        scikit-learn) to use for calculating the objective function.
        Possible values are:

        * "cityblock" for sklearn.metrics.pairwise.manhattan_distances
        * "cosine" for sklearn.metrics.pairwise.cosine_distances
        * "euclidean" for sklearn.metrics.pairwise.euclidean_distances
        * "l1" for sklearn.metrics.pairwise.manhattan_distances
        * "l2" for sklearn.metrics.pairwise.euclidean_distances
        * "manhattan" for sklearn.metrics.pairwise.manhattan_distances

        If function, then this function should take two arguments and return a
        scalar value. Furthermore, the following conditions must be fulfilled:

        1. d(a, b) >= 0, for all a and b
        2. d(a, b) == 0, if and only if a = b, positive definiteness
        3. d(a, b) == d(b, a), symmetry
        4. d(a, c) <= d(a, b) + d(b, c), the triangle inequality

    Returns
    -------
    metric_func : function
        If the `metric` argument is a function, it is returned.
        If the `metric` argument is a string, then the corresponding distance
        metric function from `sklearn.metrics.pairwise` is returned.
    """
    if metric is None:
        metric = "manhattan"

    if isinstance(metric, str):
        try:
            return distance_metrics()[metric]
        except KeyError:
            raise ValueError(
                "{} is not a known metric. Please use rather one of the "
                "following metrics: {}".format(
                    metric,
                    tuple(
                        name
                        for name in distance_metrics().keys()
                        if name != "precomputed"
                    ),
                )
            )
    elif callable(metric):
        return metric
    else:
        raise ValueError(
            "A {} was passed as `metric` argument. "
            "Please pass a string or a function "
            "instead.".format(type(metric))
        )


class MissingMetric(RuntimeError):
    """Raised when a distance metric is required but was not set."""


def raise_distance_metric_not_set(x, y):
    raise MissingMetric("distance metric not set!")


def make_move(moving_area, new_label, labels):
    """
    Modify the `labels` argument in place (no return value!) such that the
    area `moving_area` has the new region label `new_label`.

    Parameters
    ----------
    moving_area :
        The area to be moved (assigned to a new region).
    new_label : `int`
        The new region label of area `moving_area`.
    labels : :class:`numpy.ndarray`
        Each element is a region label of the area corresponding array index.

    Examples
    --------
    >>> import numpy as np
    >>> labels = np.array([0, 0, 0, 0, 1, 1])
    >>> make_move(3, 1, labels)
    >>> (labels == np.array([0, 0, 0, 1, 1, 1])).all()
    True
    """
    labels[moving_area] = new_label


def distribute_regions_among_components(component_labels, n_regions):
    r"""
    Parameters
    ----------
    component_labels : list
        Each element specifies to which connected component an area belongs.
        An example would be [0, 0, 1, 0, 0, 1] for the following two islands:
        
        ::
        
          island one        island two
          .-------.         .---.
          | 0 | 1 |         | 2 |
          | - - - |         | - |
          | 3 | 4 |         | 5 |
          `-------´         `---´

    n_regions : int

    Returns
    -------
    result_dict : Dict[int, int]
        Each key is a label of a connected component. Each value specifies into
        how many regions the component is to be clustered.
    """
    # copy list to avoid manipulating callers list instance
    component_labels = list(component_labels)
    n_regions_to_distribute = n_regions
    components = set(component_labels)
    if len(components) == 1:
        return {0: n_regions}
    result_dict = {}
    # make sure each connected component has at least one region assigned to it
    for comp in components:
        component_labels.remove(comp)
        result_dict[comp] = 1
        n_regions_to_distribute -= 1
    # distribute the rest of the regions to random components with bigger
    # components being likely to get more regions assigned to them
    while n_regions_to_distribute > 0:
        position = random.randrange(len(component_labels))
        picked_comp = component_labels.pop(position)
        result_dict[picked_comp] += 1
        n_regions_to_distribute -= 1
    return result_dict


def generate_initial_sol(adj, n_regions):
    """
    Generate a random initial clustering.

    Parameters
    ----------
    adj : :class:`scipy.sparse.csr_matrix`

    n_regions : int

    Yields
    ------
    region_labels : :class:`numpy.ndarray`
        An array with -1 for areas which are not part of the yielded
        component and an integer >= 0 specifying the region of areas within the
        yielded component.
    """
    # check args
    n_areas = adj.shape[0]
    if n_areas == 0:
        raise ValueError("There must be at least one area.")
    if n_areas < n_regions:
        raise ValueError(
            "The number of regions ({}) must be "
            "less than or equal to the number of areas "
            "({}).".format(n_regions, n_areas)
        )
    if n_regions == 1:
        yield {area: 0 for area in range(n_areas)}
        return

    n_comps, comp_labels = csg.connected_components(adj)
    if n_comps > n_regions:
        raise ValueError(
            "The number of regions ({}) must not be "
            "less than the number of connected components "
            "({}).".format(n_regions, n_comps)
        )
    n_regions_per_comp = distribute_regions_among_components(comp_labels, n_regions)

    print("n_regions_per_comp", n_regions_per_comp)
    regions_built = 0
    for comp_label, n_regions_in_comp in n_regions_per_comp.items():
        print("comp_label", comp_label)
        print("n_regions_in_comp", n_regions_in_comp)
        region_labels = -np.ones(len(comp_labels), dtype=np.int32)
        in_comp = comp_labels == comp_label
        comp_adj = adj[in_comp]
        comp_adj = comp_adj[:, in_comp]
        region_labels_comp = (
            _randomly_divide_connected_graph(comp_adj, n_regions_in_comp)
            + regions_built
        )
        regions_built += n_regions_in_comp
        print("Regions in comp:", set(region_labels_comp))
        region_labels[in_comp] = region_labels_comp
        yield region_labels


def _randomly_divide_connected_graph(adj, n_regions):
    """
    Divide the provided connected graph into `n_regions` regions.

    Parameters
    ----------
    adj : :class:`scipy.sparse.csr_matrix`
        Adjacency matrix.
    n_regions : int
        The desired number of clusters. Must be > 0 and <= number of nodes.

    Returns
    -------
    labels : :class:`numpy.ndarray`
        Each element (an integer in {0, ..., `n_regions` - 1}) specifies the
        region an area (defined by the index in the array) belongs to.

    Examples
    --------
    >>> from scipy.sparse import diags
    >>> n_nodes = 10
    >>> adj_diagonal = [1] * (n_nodes-1)
    >>> # 10x10 adjacency matrix representing the path 0-1-2-...-9-10
    >>> adj = diags([adj_diagonal, adj_diagonal], offsets=[-1, 1])
    >>> n_regions_desired = 4
    >>> labels = _randomly_divide_connected_graph(adj, n_regions_desired)
    >>> n_regions_obtained = len(set(labels))
    >>> n_regions_desired == n_regions_obtained
    True
    """
    if not n_regions > 0:
        msg = "n_regions is {} but must be positive.".format(n_regions)
        raise ValueError(msg)
    n_areas = adj.shape[0]
    if not n_regions <= n_areas:
        msg = (
            "n_regions is {} but must less than or equal to "
            + "the number of nodes which is {}".format(n_regions, n_areas)
        )
        raise ValueError(msg)
    mst = csg.minimum_spanning_tree(adj)
    for _ in range(n_regions - 1):
        # try different links to cut and pick the one leading to the most
        # balanced solution
        best_link = None
        max_region_size = float("inf")
        for __ in range(5):
            mst_copy = mst.copy()
            nonzero_i, nonzero_j = mst_copy.nonzero()
            random_position = random.randrange(len(nonzero_i))
            i, j = nonzero_i[random_position], nonzero_j[random_position]
            mst_copy[i, j] = 0
            mst_copy.eliminate_zeros()
            labels = csg.connected_components(mst_copy, directed=False)[1]
            max_size = max(np.unique(labels, return_counts=True)[1])
            if max_size < max_region_size:
                best_link = (i, j)
                max_region_size = max_size
        mst[best_link[0], best_link[1]] = 0
        mst.eliminate_zeros()
    return csg.connected_components(mst)[1]


def copy_func(f):
    """
    Return a copy of a function. This is useful e.g. to create aliases (whose
    docstrings can be changed without affecting the original function).
    The implementation is taken from https://stackoverflow.com/a/13503277.

    Parameters
    ----------
    f : function

    Returns
    -------
    g : function
        Copy of `f`.
    """
    g = types.FunctionType(
        f.__code__,
        f.__globals__,
        name=f.__name__,
        argdefs=f.__defaults__,
        closure=f.__closure__,
    )
    g = functools.update_wrapper(g, f)
    g.__kwdefaults__ = f.__kwdefaults__
    return g


def assert_feasible(solution, adj, n_regions=None):
    """
    Parameters
    ----------
    solution : :class:`numpy.ndarray`
        Array of region labels.
    adj : :class:`scipy.sparse.csr_matrix`
        Adjacency matrix representing the contiguity relation.
    n_regions : `int` or `None`
        An `int` represents the desired number of regions.
        If `None`, then the number of regions is not checked.

    Raises
    ------
    exc : `ValueError`
        A `ValueError` is raised if clustering is not spatially contiguous.
        Given the `n_regions` argument is not `None`, a `ValueError` is raised
        also if the number of regions is not equal to the `n_regions` argument.
    """
    if n_regions is not None:
        if len(set(solution)) != n_regions:
            raise ValueError(
                "The number of regions is {} but "
                "should be {}".format(len(solution), n_regions)
            )

    for region_label in set(solution):
        aux = sub_adj_matrix(adj, np.where(solution == region_label)[0])

        # check right contiguity
        if not is_connected(aux):
            raise ValueError(
                "Region {} is not spatially " "contiguous.".format(region_label)
            )


def boolean_assert_feasible(solution, adj, n_regions=None):
    """
    Return boolean version of assert_feasible
    """

    resp = []
    if n_regions is not None:
        if len(set(solution)) != n_regions:
            raise ValueError(
                "The number of regions is {} but "
                "should be {}".format(len(solution), n_regions)
            )

    for region_label in set(solution):
        aux = sub_adj_matrix(adj, np.where(solution == region_label)[0])
        resp.append(is_connected(aux))

    final_resp = all(resp)
    return final_resp


def all_elements_equal(array):
    return np.max(array) == np.min(array)


def separate_components(adj, labels):
    """
    Take a labels array and yield modifications of it (one modified array per
    connected component). The modified array will be unchanged at those indices
    belonging to the current connected component. Thus it will have integers
    >= 0 there. At all other indices the Yielded array will be -1.

    Parameters
    ----------
    adj : :class:`scipy.sparse.csr_matrix`
        Adjacency matrix representing the contiguity relation.
    labels : :class:`numpy.ndarray`

    Yields
    ------
    comp_dict : :class:`numpy.ndarray`
        Each yielded dict represents one connected component of the graph
        specified by the `adj` argument. In a yielded dict, each key is an area
        and each value is the corresponding region-ID.

    Examples
    --------
    >>> edges_island1 = [(0, 1), (1, 2),          # 0 | 1 | 2
    ...                  (0, 3), (1, 4), (2, 5),  # ---------
    ...                  (3, 4), (4,5)]           # 3 | 4 | 5
    >>>
    >>> edges_island2 = [(6, 7),                  # 6 | 7
    ...                  (6, 8), (7, 9),          # -----
    ...                  (8, 9)]                  # 8 | 9
    >>>
    >>> graph = nx.Graph(edges_island1 + edges_island2)
    >>> adj = nx.to_scipy_sparse_matrix(graph)
    >>>
    >>> # island 1: island divided into regions 0, 1, and 2
    >>> sol_island1 = [area%3 for area in range(6)]
    >>> # island 2: all areas are in region 3
    >>> sol_island2 = [3 for area in range(6, 10)]
    >>> labels = np.array(sol_island1 + sol_island2)
    >>>
    >>> yielded = list(separate_components(adj, labels))
    >>> yielded.sort(key=lambda arr: arr[0], reverse=True)
    >>> (yielded[0] == np.array([0, 1, 2, 0, 1, 2, -1, -1, -1, -1])).all()
    True
    >>> (yielded[1] == np.array([-1, -1, -1, -1, -1, -1, 3, 3, 3, 3])).all()
    True
    """
    n_comps, comp_labels = csg.connected_components(adj)
    for comp in set(comp_labels):
        region_labels = -np.ones(len(comp_labels), dtype=np.int32)
        in_comp = comp_labels == comp
        region_labels[in_comp] = labels[in_comp]
        yield region_labels


def random_element_from(lst):
    random_position = random.randrange(len(lst))
    return lst[random_position]


def pop_randomly_from(lst):
    random_position = random.randrange(len(lst))
    return lst.pop(random_position)


def count(arr, el):
    """
    Parameters
    ----------
    arr : :class:`numpy.ndarray`

    el : object

    Returns
    -------
    result : :class:`numpy.ndarray`
        The number of occurences of `el` in `arr`.

    Examples
    --------
    >>> arr = np.array([0, 0, 0, 1, 1])
    >>> count(arr, 0)
    3
    >>> count(arr, 1)
    2
    >>> count(arr, 2)
    0
    """
    unique, counts = np.unique(arr, return_counts=True)
    idx = np.where(unique == el)[0]
    if len(idx) > 0:
        return int(counts[idx])
    return 0


def check_solver(solver):
    if not isinstance(solver, str) or solver.lower() not in [
        "cbc",
        "cplex",
        "glpk",
        "gurobi",
    ]:
        raise ValueError(
            "The solver argument must be one of the following"
            ' strings: "cbc", "cplex", "glpk", or "gurobi".'
        )


def get_solver_instance(solver_string):
    solver = {
        "cbc": pulp.solvers.PULP_CBC_CMD,
        "cplex": pulp.solvers.CPLEX,
        "glpk": pulp.solvers.GLPK,
        "gurobi": pulp.solvers.GUROBI,
    }[solver_string.lower()]
    solver_instance = solver()
    return solver_instance
